"""
Generate sample x, y, data of hierarchical structure for bebi103
documentation.
"""

import numpy as np
import pandas as pd


def _generate_sample_data():
    np.random.seed(3252)
    J_1 = 3
    n = np.array([20, 25, 18])
    theta = np.array([3, 7])
    tau = np.array([1, 4])
    sigma = np.array([2, 3])
    rho = 0.6

    sigma_cov = np.array(
        [
            [sigma[0] ** 2, rho * sigma[0] * sigma[1]],
            [rho * sigma[0] * sigma[1], sigma[1] ** 2],
        ]
    )

    x = []
    y = []
    trial = []

    for i, n_val in enumerate(n):
        theta_1 = np.random.multivariate_normal(theta, np.diag(tau ** 2))
        x_vals, y_vals = np.random.multivariate_normal(
            theta_1, sigma_cov, size=n_val
        ).transpose()
        x += list(x_vals)
        y += list(y_vals)
        trial += [i + 1] * n_val

    return pd.DataFrame(dict(x=x, y=y, trial=trial))


if __name__ == "__main__":
    df = _generate_sample_data()
    df.to_csv('sample_data.csv', index=False)



import os

import multiprocessing
import warnings

import tqdm

import numpy as np
import numba

def unpivot_csv(
    input_file: str,
    output_file: str,
    n_header_rows=None,
    header_names=None,
    value_name: str = "value",
    sep: str = ",",
    retain_row_index: bool = True,
    row_index_name: str = "original_row_index",
    comment_prefix=None,
    force_overwrite=False,
    include_comments=True,
    additional_comment=None,
) -> type(None):
    """Converts a CSV file in wide format, possibly with a hierarchical
    column, index to tall format via a unpivot operation.

    Parameters
    ----------
    input_file : str
        Path to CSV file containing data in a wide format.
    output_file : str
        Path to file to write contents in tall format.
    n_header_rows : int, default 1
        Number of header rows, ignored if `header_names` is given.
    header_names : string, list of strings, or None, default None
        Names of each level of the header. These become the names of
        columns in the tall representation of the data set. If None,
        column names are variable_0, variable_1, etc.
    value_name : str or None, default 'value'
        Name to assign to column containing values in tall
        representation.
    sep : str, default ','
        Separator. Cannot be a regex; must be a single character.
    retain_row_index : bool, default True
        If True, include a column containing the original row index
        in the original file. This is useful to avoid information loss,
        since all entries in a given row can be related. These are 
        essentially identifies variables.
    row_index_name : str, default 'original_row_index'
        Label in the row index if the row index is retained. Ignored if
        `retain_row_index` is False.
    comment_prefix : str or None, default None
        A single character defining which rows in the input file should
        be treated as comments.
    force_overwrite : bool, default False
        If True, if the output_file already exists, it is overwritten.
    include_comments : bool, default True
        If True, write comments from the input file in the output files.
    additional_comment : str or None, default None
        Additional comment at top of output file.

    Returns
    -------
    output : None
        Does not return anything. The output is written to output_file.
    """
    if input_file == output_file:
        raise RuntimeError("`input_file` and `output_file` are the same.")

    if not force_overwrite and os.path.isfile(output_file):
        raise RuntimeError(f"File {output_file} exists, not overwriting.")

    if header_names is None:
        if n_header_rows is None:
            n_header_rows = 1
        header_names = [f"variable_{i}" for i in range(n_header_rows)]
    elif np.isscalar(header_names):
        header_names = [header_names]

    if n_header_rows is None:
        n_header_rows = len(header_names)

    if len(header_names) != n_header_rows:
        raise RuntimeError("len(header_names) != n_header_rows")

    with open(input_file, "r") as f, open(output_file, "w") as fout:
        # Read and write comments
        line = f.readline()
        while line != "" and line[0] == comment_prefix:
            if include_comments:
                fout.write(line)
            line = f.readline()

        # Headers, a list of lists
        headers = [line.rstrip().split(sep)] + [
            f.readline().rstrip().split(sep) for _ in range(n_header_rows - 1)
        ]

        # Header to output
        if retain_row_index:
            fout.write(f"{row_index_name},")
        fout.write(sep.join(header_names + [value_name]) + "\n")

        # Now read data and add to output file
        row_index = 0
        line = f.readline()
        while line != "":
            data = line.split(sep)
            for i, data_str in enumerate(data):
                data_str = data_str.strip().rstrip()
                if retain_row_index:
                    fout.write(str(row_index) + sep)
                if data_str != "":
                    fout.write(
                        sep.join([head[i] for head in headers] + [data_str]) + "\n"
                    )
            line = f.readline()
            row_index += 1

    return None


def _dummy_jit(*args, **kwargs):
    """Dummy wrapper for jitting if numba not applicable."""

    def wrapper(f):
        return f

    def marker(*args, **kwargs):
        return marker

    if (
        len(args) > 0
        and (args[0] is marker or not callable(args[0]))
        or len(kwargs) > 0
    ):
        return wrapper
    elif len(args) == 0:
        return wrapper
    else:
        return args[0]


@numba.jit(nopython=True)
def _ecdf(x, data):
    """
    Compute the values of the formal ECDF generated from `data` at
    points `x`. I.e., if F is the ECDF, return F(x).

    Parameters
    ----------
    x : array_like
        Positions at which the formal ECDF is to be evaluated.
    data : array_like
        Data set to use to generate the ECDF.

    Returns
    -------
    output : float or ndarray
        Value of the ECDF at `x`.
    """
    y = np.arange(len(data) + 1) / len(data)
    return y[np.searchsorted(np.sort(data), x, side="right")]


@numba.jit(nopython=True)
def _pearson_r(x, y):
    """
    Compute the Pearson correlation coefficient between two samples.

    Parameters
    ----------
    data_1 : array_like
        One-dimensional array of data.
    data_2 : array_like
        One-dimensional array of data.

    Returns
    -------
    output : float
        The Pearson correlation coefficient between `data_1`
        and `data_2`.

    Notes
    -----
    Only entries where both `data_1` and `data_2` are not NaN are
    used.

    If the variance of `data_1` or `data_2` is zero, return NaN.
    """
    if _allequal(x) or _allequal(y):
        return np.nan

    return (np.mean(x * y) - np.mean(x) * np.mean(y)) / np.std(x) / np.std(y)


def _convert_data(data, inf_ok=False, min_len=1):
    """
    Convert inputted 1D data set into NumPy array of floats.
    All nan's are dropped.

    Parameters
    ----------
    data : int, float, or array_like
        Input data, to be converted.
    inf_ok : bool, default False
        If True, np.inf values are allowed in the arrays.
    min_len : int, default 1
        Minimum length of array.

    Returns
    -------
    output : ndarray
        `data` as a one-dimensional NumPy array, dtype float.
    """
    # If it's scalar, convert to array
    if np.isscalar(data):
        data = np.array([data], dtype=float)

    # Convert data to NumPy array
    data = np.array(data, dtype=float)

    # Make sure it is 1D
    if len(data.shape) != 1:
        raise RuntimeError("Input must be a 1D array or Pandas series.")

    # Remove NaNs
    data = data[~np.isnan(data)]

    # Check for infinite entries
    if not inf_ok and np.isinf(data).any():
        raise RuntimeError("All entries must be finite.")

    # Check to minimal length
    if len(data) < min_len:
        raise RuntimeError(
            "Array must have at least {0:d} non-NaN entries.".format(min_len)
        )

    return data


def _convert_two_data(x, y, inf_ok=False, min_len=1):
    """
    Converted two inputted 1D data sets into Numpy arrays of floats.
    Indices where one of the two arrays is nan are dropped.

    Parameters
    ----------
    x : array_like
        Input data, to be converted. `x` and `y` must have the same length.
    y : array_like
        Input data, to be converted. `x` and `y` must have the same length.
    inf_ok : bool, default False
        If True, np.inf values are allowed in the arrays.
    min_len : int, default 1
        Minimum length of array.

    Returns
    -------
    x_out : ndarray
        `x` as a one-dimensional NumPy array, dtype float.
    y_out : ndarray
        `y` as a one-dimensional NumPy array, dtype float.
    """
    # Make sure they are array-like
    if np.isscalar(x) or np.isscalar(y):
        raise RuntimeError("Arrays must be 1D arrays of the same length.")

    # Convert to Numpy arrays
    x = np.array(x, dtype=float)
    y = np.array(y, dtype=float)

    # Check for infinite entries
    if not inf_ok and (np.isinf(x).any() or np.isinf(y).any()):
        raise RuntimeError("All entries in arrays must be finite.")

    # Make sure they are 1D arrays
    if len(x.shape) != 1 or len(y.shape) != 1:
        raise RuntimeError("Input must be a 1D array or Pandas series.")

    # Must be the same length
    if len(x) != len(y):
        raise RuntimeError("Arrays must be 1D arrays of the same length.")

    # Clean out nans
    inds = ~np.logical_or(np.isnan(x), np.isnan(y))
    x = x[inds]
    y = y[inds]

    # Check to minimal length
    if len(x) < min_len:
        raise RuntimeError(
            "Arrays must have at least {0:d} mutual non-NaN entries.".format(min_len)
        )

    return x, y


@numba.jit(nopython=True)
def _allequal(x, rtol=1e-7, atol=1e-14):
    """
    Determine if all entries in an array are equal.

    Parameters
    ----------
    x : ndarray
        Array to test.

    Returns
    -------
    output : bool
        True is all entries in the array are equal, False otherwise.
    """
    if len(x) == 1:
        return True

    for a in x[1:]:
        if np.abs(a - x[0]) > (atol + rtol * np.abs(a)):
            return False
    return True


@numba.jit(nopython=True)
def _allclose(x, y, rtol=1e-7, atol=1e-14):
    """
    Determine if all entries in two arrays are close to each other.

    Parameters
    ----------
    x : ndarray
        First array to compare.
    y : ndarray
        Second array to compare.

    Returns
    -------
    output : bool
        True is each entry in the respective arrays is equal.
        False otherwise.
    """
    for a, b in zip(x, y):
        if np.abs(a - b) > (atol + rtol * np.abs(b)):
            return False
    return True


def _make_one_arg_numba_func(func, func_args):
    """
    Make a Numba'd version of a function that takes one positional
    argument.

    Parameters
    ----------
    func : function
        Function with call signature `func(x, *args)`.
    func_args : tuple
        Tuple of args to use in testing a function call.

    Returns
    -------
    output : Numba'd function (or original function)
        A Numba'd version of the function. If that is not possible,
        returns the original function.
    numba_success : bool
        True if function was successfully jitted.

    """
    try:
        func_numba = numba.jit(func, nopython=True)

        @numba.jit(nopython=True)
        def f(x, args=()):
            return func_numba(x, *args)

        # Attempt function call
        _ = f(np.array([1.0, 2.0]), func_args)

        return f, True
    except:

        def f(x, args=()):
            return func(x, *args)

        return f, False


def _make_two_arg_numba_func(func, func_args):
    """
    Make a Numba'd version of a function that takes two positional
    arguments.

    Parameters
    ----------
    func : function
        Function with call signature `func(x, y, *args)`.
    func_args : tuple
        Tuple of args to use in testing a function call.

    Returns
    -------
    output : Numba'd function (or original function)
        A Numba'd version of the function. If that is not possible,
        returns the original function.
    numba_success : bool
        True if function was successfully jitted.
    """
    try:
        func_numba = numba.jit(func, nopython=True)

        @numba.jit(nopython=True)
        def f(x, args=()):
            return func_numba(x, *args)

        # Attempt function call
        _ = f(np.array([1.0, 2.0]), np.array([1.0, 2.0]), func_args)

        return f, True
    except:

        def f(x, y, args=()):
            return func(x, y, *args)

        return f, False


@numba.jit(nopython=True)
def _seed_numba(seed):
    """
    Seed the random number generator for Numba'd functions.

    Parameters
    ----------
    seed : int
        Seed of the RNG.

    Returns
    -------
    None
    """
    np.random.seed(seed)



import numpy as np
import pandas as pd

import bokeh.plotting

import bebi103


def test_bokeh_boxplot():
    labels = list('ADBACBABDABDBDBCBABCBADBADBBADBACBADBAD')
    values = np.array([1.67,  0.23,  2.51,  1.06,  0.22,
                      -0.54,  1.07,  0.07, -0.15, -1.47,
                       1.42, -0.28, -0.38, -1.24, -0.78,
                      -0.25, -0.49,  0.78,  0.25,  0.81,
                      -1.20, -0.53, -1.33, -0.36, -0.63,
                       0.74, -0.77, -0.68,  0.04,  0.32,
                       0.35,  0.90, -2.91,  1.04,  0.86,
                      -0.76,  0.56,  1.98, -1.13])
    df = pd.DataFrame({'label': labels, 'value': values})
    bokeh.plotting.output_file('test_boxplot.html')
    p = bebi103.bokeh_boxplot(df, value='value', label='label', sort=False)
    bokeh.plotting.show(p)
    print('Medians:')
    print(df.groupby('label').median())
    y_n = input('Visual inspection ok? ')
    if y_n in ['y', 'Y', '1']:
        return True
    else:
        return False


def test_bokeh_matplot():
    a = np.array([[1, 2, 3], [4, 5, 6], [7, 8, None]]).astype(float)
    a[1, 1] = None
    data = np.array(np.unravel_index(range(9), a.shape) + (a.ravel(),)).T
    df = pd.DataFrame(data, columns=['i', 'j', 'data'])

    bokeh.plotting.output_file('test_matplot.html')
    # Numerical x-axis
    p_numerical = bebi103.bokeh_matplot(df, 'i', 'j', 'data', n_colors=9,
                                        colormap='RdBu_r', plot_width=200,
                                        plot_height=200)

    # Categorical x-axis, i.e. column indices (labels are strings)
    df.j = df.j.astype(str)
    p_categorical = bebi103.bokeh_matplot(df, 'i', 'j', 'data', n_colors=9,
                                          colormap='RdBu_r', plot_width=200,
                                          plot_height=200)

    bokeh.plotting.show(bokeh.io.vplot(p_categorical, p_numerical))
    y_n = input('Visual inspection ok? (X-Axis -> Top: above, Bottom: below)')
    if y_n in ['y', 'Y', '1']:
        return True
    else:
        return False


# To go in run_ensemble_sampler:
# assert np.allclose(sampler.chain[0,:,0], df[df.chain==0][df.columns[0]])
# assert np.allclose(sampler.chain[1,:,0], df[df.chain==1][df.columns[0]])
# assert np.allclose(sampler.chain[13,:,1], df[df.chain==13][df.columns[1]])
# assert np.allclose(sampler.chain[45,:,1], df[df.chain==45][df.columns[1]])

# To go in run_pt_sampler:
# inds = (df['chain'] == 0) & (df['beta_ind'] == 2)
# assert np.allclose(sampler.chain[2, 0, :, 0], df[inds][df.columns[0]])
# assert np.allclose(sampler.lnprobability[2, 0], df[inds]['lnprob'])



def bokeh_matplot(df, i_col, j_col, data_col, data_range=None, n_colors=21,
                  label_ticks=True, colormap='RdBu_r', plot_width=1000,
                  plot_height=1000, x_axis_location='auto',
                  toolbar_location='left',
                  tools='reset,resize,hover,save,pan,box_zoom,wheel_zoom',
                  **kwargs):
    """
    Create Bokeh plot of a matrix.

    Parameters
    ----------
    df : Pandas DataFrame
        Tidy DataFrame to be plotted as a matrix.
    i_col : hashable object
        Column in `df` to be used for row indices of matrix.
    j_col : hashable object
        Column in `df` to be used for column indices of matrix.
    data_col : hashable object
        Column containing values to be plotted.  These values
        set which color is displayed in the plot and also are
        displayed in the hover tool.
    data_range : array_like, shape (2,)
        Low and high values that data may take, used for scaling
        the color.  Default is the range of the inputted data.
    n_colors : int, default = 21
        Number of colors to be used in colormap.
    label_ticks : bool, default = True
        If False, do not put tick labels
    colormap : str, default = 'RdBu_r'
        Any of the allowed seaborn colormaps.
    plot_width : int, default 1000
        Width of plot in pixels.
    plot_height : int, default 1000
        Height of plot in pixels.
    x_axis_location : str, default = None
        Location of the x-axis around the plot. If 'auto' and first
        element of `df[i_col]` is numerical, x-axis will be placed below
        with the lower left corner as the origin. Otherwise, above
        with the upper left corner as the origin.
    toolbar_location : str, default = 'left'
        Location of the Bokeh toolbar around the plot
    tools : str, default = 'reset,resize,hover,save,pan,box_zoom,wheel_zoom'
        Tools to show in the Bokeh toolbar
    **kwargs
        Arbitrary keyword arguments passed to bokeh.plotting.figure

    Returns
    -------
    Bokeh plotting object


    Examples
    --------
    >>> a = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> data = np.array(np.unravel_index(range(9), a.shape) + (a.ravel(),)).T
    >>> df = pd.DataFrame(data, columns=['i', 'j', 'data'])
    >>> bokeh.plotting.output_file('test_matplot.html')
    >>> p = bokeh_matplot(df, i_col, j_col, data_col, n_colors=21,
                          colormap='RdBu_r', plot_width=1000,
                          plot_height=1000)
    >>> bokeh.plotting.show(p)
    """
    # Copy the DataFrame
    df_ = df.copy()

    # Convert i, j to strings so not interpreted as physical space
    df_[i_col] = df_[i_col].astype(str)
    df_[j_col] = df_[j_col].astype(str)

    # Get data range
    if data_range is None:
        data_range = (df[data_col].min(), df[data_col].max())
    elif (data_range[0] > df[data_col].min()) \
            or (data_range[1] < df[data_col].max()):
        raise RuntimeError('Data out of specified range.')

    # Get colors
    palette = sns.color_palette(colormap, n_colors)

    # Compute colors for squares
    df_['color'] = df_[data_col].apply(data_to_hex_color,
                                       args=(palette, data_range))

    # Data source
    source = bokeh.plotting.ColumnDataSource(df_)

    # only reverse the y-axis and put the x-axis on top
    # if the x-axis is categorical:
    if x_axis_location == 'auto':
        if isinstance(df[j_col].iloc[0], numbers.Number):
            y_range = list(df_[i_col].unique())
            x_axis_location = 'below'
        else:
            y_range = list(reversed(list(df_[i_col].unique())))
            x_axis_location = 'above'
    elif x_axis_location == 'above':
        y_range = list(reversed(list(df_[i_col].unique())))
    elif x_axis_location == 'below':
        y_range = list(df_[i_col].unique())

    # Set up figure
    p = bokeh.plotting.figure(x_range=list(df_[j_col].unique()),
                              y_range=y_range,
                              x_axis_location=x_axis_location,
                              plot_width=plot_width,
                              plot_height=plot_height,
                              toolbar_location=toolbar_location,
                              tools=tools, **kwargs)

    # Populate colored squares
    p.rect(j_col, i_col, 1, 1, source=source, color='color', line_color=None)

    # Set remaining properties
    p.grid.grid_line_color = None
    p.axis.axis_line_color = None
    p.axis.major_tick_line_color = None
    if label_ticks:
        p.axis.major_label_text_font_size = '8pt'
    else:
        p.axis.major_label_text_color = None
        p.axis.major_label_text_font_size = '0pt'
    p.axis.major_label_standoff = 0
    p.xaxis.major_label_orientation = np.pi / 3

    # Build hover tool
    hover = p.select(dict(type=bokeh.models.HoverTool))
    hover.tooltips = collections.OrderedDict([('i', '  @' + i_col),
                                              ('j', '  @' + j_col),
                                              (data_col, '  @' + data_col)])

    return p


def bokeh_boxplot(df, value, label, ylabel=None, sort=True, plot_width=650,
                  plot_height=450, box_fill_color='medium_purple',
                  background_fill_color='#DFDFE5',
                  tools='reset,resize,hover,save,pan,box_zoom,wheel_zoom',
                  **kwargs):
    """
    Make a Bokeh box plot from a tidy DataFrame.

    Parameters
    ----------
    df : tidy Pandas DataFrame
        DataFrame to be used for plotting
    value : hashable object
        Column of DataFrame containing data to be used.
    label : hashable object
        Column of DataFrame use to categorize.
    ylabel : str, default None
        Text for y-axis label
    sort : Boolean, default True
        If True, sort DataFrame by label so that x-axis labels are
        alphabetical.
    plot_width : int, default 650
        Width of plot in pixels.
    plot_height : int, default 450
        Height of plot in pixels.
    box_fill_color : string
        Fill color of boxes, default = 'medium_purple'
    background_fill_color : str, default = '#DFDFE5'
        Fill color of the plot background
    tools : str, default = 'reset,resize,hover,save,pan,box_zoom,wheel_zoom'
        Tools to show in the Bokeh toolbar
    **kwargs
        Arbitrary keyword arguments passed to bokeh.plotting.figure

    Returns
    -------
    Bokeh plotting object

    Example
    -------
    >>> cats = list('ABCD')
    >>> values = np.random.randn(200)
    >>> labels = np.random.choice(cats, 200)
    >>> df = pd.DataFrame({'label': labels, 'value': values})
    >>> bokeh.plotting.output_file('test_boxplot.html')
    >>> p = bokeh_boxplot(df, value='value', label='label')
    >>> bokeh.plotting.show(p)

    Notes
    -----
    .. Based largely on example code found here:
     https://github.com/bokeh/bokeh/blob/master/examples/plotting/file/boxplot.py
    """

    # Sort DataFrame by labels for alphabetical x-labeling
    if sort:
        df_sort = df.sort_values(label)
    else:
        df_sort = df.copy()

    # Convert labels to string to allow categorical axis labels
    df_sort[label] = df_sort[label].astype(str)

    # Get the categories
    cats = list(df_sort[label].unique())

    # Group Data frame
    df_gb = df_sort.groupby(label)

    # Compute quartiles for each group
    q1 = df_gb[value].quantile(q=0.25)
    q2 = df_gb[value].quantile(q=0.5)
    q3 = df_gb[value].quantile(q=0.75)

    # Compute interquartile region and upper and lower bounds for outliers
    iqr = q3 - q1
    upper_cutoff = q3 + 1.5 * iqr
    lower_cutoff = q1 - 1.5 * iqr

    # Find the outliers for each category
    def outliers(group):
        cat = group.name
        outlier_inds = (group[value] > upper_cutoff[cat]) | \
                       (group[value] < lower_cutoff[cat])
        return group[value][outlier_inds]

    # Apply outlier finder
    out = df_gb.apply(outliers).dropna()

    # Points of outliers for plotting
    outx = []
    outy = []
    if not out.empty:
        for cat in cats:
            if not out[cat].empty:
                for val in out[cat]:
                    outx.append(cat)
                    outy.append(val)

    # Shrink whiskers to smallest and largest non-outlier
    qmin = df_gb[value].min()
    qmax = df_gb[value].max()
    upper = upper_cutoff.combine(qmax, min)
    lower = lower_cutoff.combine(qmin, max)

    # Reindex to make sure ordering is right when plotting
    upper = upper.reindex(cats)
    lower = lower.reindex(cats)
    q1 = q1.reindex(cats)
    q2 = q2.reindex(cats)
    q3 = q3.reindex(cats)

    # Build figure
    p = bokeh.plotting.figure(x_range=cats,
                              background_fill_color=background_fill_color,
                              plot_width=plot_width, plot_height=plot_height,
                              tools=tools,
                              **kwargs)
    p.ygrid.grid_line_color = 'white'
    p.xgrid.grid_line_color = None
    p.ygrid.grid_line_width = 2
    p.yaxis.axis_label = ylabel

    # stems
    p.segment(cats, upper, cats, q3, line_width=2, line_color="black")
    p.segment(cats, lower, cats, q1, line_width=2, line_color="black")

    # boxes
    p.rect(cats, (q3 + q1) / 2, 0.5, q3 - q1, fill_color="mediumpurple",
           alpha=0.7, line_width=2, line_color="black")

    # median (almost-0 height rects simpler than segments)
    y_range = qmax.max() - qmin.min()
    p.rect(cats, q2, 0.5, 0.0001 * y_range, line_color="black",
           line_width=2, fill_color='black')

    # whiskers (almost-0 height rects simpler than segments with
    # categorial x-axis)
    p.rect(cats, lower, 0.2, 0.0001 * y_range, line_color='black',
           fill_color='black')
    p.rect(cats, upper, 0.2, 0.0001 * y_range, line_color='black',
           fill_color='black')

    # outliers
    p.circle(outx, outy, size=6, color='black')

    return p


def _outliers(data):
    bottom, middle, top = np.percentile(data, [25, 50, 75])
    iqr = top - bottom
    outliers = data[(data > top + 1.5 * iqr) | (data < bottom - 1.5 * iqr)]
    return outliers


def _box_and_whisker(data):
    middle = data.median()
    bottom = data.quantile(0.25)
    top = data.quantile(0.75)
    iqr = top - bottom
    top_whisker = data[data <= top + 1.5 * iqr].max()
    bottom_whisker = data[data >= bottom - 1.5 * iqr].min()
    return pd.Series(
        {
            "middle": middle,
            "bottom": bottom,
            "top": top,
            "top_whisker": top_whisker,
            "bottom_whisker": bottom_whisker,
        }
    )


def _box_source(df, cats, val, cols):
    """Construct a data frame for making box plot."""

    # Need to reset index for use in slicing outliers
    df_source = df.reset_index(drop=True)

    if type(cats) in [list, tuple]:
        level = list(range(len(cats)))
    else:
        level = 0

    if cats is None:
        grouped = df_source
    else:
        grouped = df_source.groupby(cats)

    # Data frame for boxes and whiskers
    df_box = grouped[val].apply(_box_and_whisker).unstack().reset_index()
    source_box = _cat_source(
        df_box, cats, ["middle", "bottom", "top", "top_whisker", "bottom_whisker"], None
    )

    # Data frame for outliers
    df_outliers = grouped[val].apply(_outliers).reset_index(level=level)
    df_outliers[cols] = df_source.loc[df_outliers.index, cols]
    source_outliers = _cat_source(df_outliers, cats, cols, None)

    return source_box, source_outliers


def _check_cat_input(df, cats, val, color_column, tooltips, palette, kwargs):
    if df is None:
        raise RuntimeError("`df` argument must be provided.")
    if cats is None:
        raise RuntimeError("`cats` argument must be provided.")
    if val is None:
        raise RuntimeError("`val` argument must be provided.")

    if type(palette) not in [list, tuple]:
        raise RuntimeError("`palette` must be a list or tuple.")

    if val not in df.columns:
        raise RuntimeError(f"{val} is not a column in the inputted data frame")

    cats_array = type(cats) in [list, tuple]

    if cats_array:
        for cat in cats:
            if cat not in df.columns:
                raise RuntimeError(f"{cat} is not a column in the inputted data frame")
    else:
        if cats not in df.columns:
            raise RuntimeError(f"{cats} is not a column in the inputted data frame")

    if color_column is not None and color_column not in df.columns:
        raise RuntimeError(f"{color_column} is not a column in the inputted data frame")

    cols = _cols_to_keep(cats, val, color_column, tooltips)

    for col in cols:
        if col not in df.columns:
            raise RuntimeError(f"{col} is not a column in the inputted data frame")

    bad_kwargs = ["x", "y", "source", "cat", "legend"]
    if kwargs is not None and any([key in kwargs for key in bad_kwargs]):
        raise RuntimeError(", ".join(bad_kwargs) + " are not allowed kwargs.")

    if val == "cat":
        raise RuntimeError("`'cat'` cannot be used as `val`.")

    if val == "__label" or (cats == "__label" or (cats_array and "__label" in cats)):
        raise RuntimeError("'__label' cannot be used for `val` or `cats`.")

    return cols

def _get_cat_range(df, grouped, order, color_column, horizontal):
    if order is None:
        if isinstance(list(grouped.groups.keys())[0], tuple):
            factors = tuple(
                [tuple([str(k) for k in key]) for key in grouped.groups.keys()]
            )
        else:
            factors = tuple([str(key) for key in grouped.groups.keys()])
    else:
        if type(order[0]) in [list, tuple]:
            factors = tuple([tuple([str(k) for k in key]) for key in order])
        else:
            factors = tuple([str(entry) for entry in order])

    if horizontal:
        cat_range = bokeh.models.FactorRange(*(factors[::-1]))
    else:
        cat_range = bokeh.models.FactorRange(*factors)

    if color_column is None:
        color_factors = factors
    else:
        color_factors = tuple(sorted(list(df[color_column].unique().astype(str))))

    return cat_range, factors, color_factors


def _cat_figure(
    df,
    grouped,
    plot_height,
    plot_width,
    x_axis_label,
    y_axis_label,
    title,
    order,
    color_column,
    tooltips,
    horizontal,
    val_axis_type,
):
    fig_kwargs = dict(
        plot_height=plot_height,
        plot_width=plot_width,
        x_axis_label=x_axis_label,
        y_axis_label=y_axis_label,
        title=title,
        tooltips=tooltips,
    )

    cat_range, factors, color_factors = _get_cat_range(
        df, grouped, order, color_column, horizontal
    )

    if horizontal:
        fig_kwargs["y_range"] = cat_range
        fig_kwargs["x_axis_type"] = val_axis_type
    else:
        fig_kwargs["x_range"] = cat_range
        fig_kwargs["y_axis_type"] = val_axis_type

    return bokeh.plotting.figure(**fig_kwargs), factors, color_factors


def _cat_source(df, cats, cols, color_column):
    if type(cats) in [list, tuple]:
        cat_source = list(zip(*tuple([df[cat].astype(str) for cat in cats])))
        labels = [", ".join(cat) for cat in cat_source]
    else:
        cat_source = list(df[cats].astype(str).values)
        labels = cat_source

    if type(cols) in [list, tuple, pd.core.indexes.base.Index]:
        source_dict = {col: list(df[col].values) for col in cols}
    else:
        source_dict = {cols: list(df[cols].values)}

    source_dict["cat"] = cat_source
    if color_column in [None, "cat"]:
        source_dict["__label"] = labels
    else:
        source_dict["__label"] = list(df[color_column].astype(str).values)
        source_dict[color_column] = list(df[color_column].astype(str).values)

    return bokeh.models.ColumnDataSource(source_dict)


def _tooltip_cols(tooltips):
    if tooltips is None:
        return []
    if type(tooltips) not in [list, tuple]:
        raise RuntimeError("`tooltips` must be a list or tuple of two-tuples.")

    cols = []
    for tip in tooltips:
        if type(tip) not in [list, tuple] or len(tip) != 2:
            raise RuntimeError("Invalid tooltip.")
        if tip[1][0] == "@":
            if tip[1][1] == "{":
                cols.append(tip[1][2 : tip[1].find("}")])
            elif "{" in tip[1]:
                cols.append(tip[1][1 : tip[1].find("{")])
            else:
                cols.append(tip[1][1:])

    return cols


def _cols_to_keep(cats, val, color_column, tooltips):
    cols = _tooltip_cols(tooltips)
    cols += [val]

    if type(cats) in [list, tuple]:
        cols += list(cats)
    else:
        cols += [cats]

    if color_column is not None:
        cols += [color_column]

    return list(set(cols))

def _point_ecdf_source(data, val, cats, cols, complementary, colored):
    """DataFrame for making point-wise ECDF."""
    df = data.copy()

    if complementary:
        col = "__ECCDF"
    else:
        col = "__ECDF"

    if cats is None or colored:
        df[col] = _ecdf_y(df[val], complementary)
    else:
        df[col] = df.groupby(cats)[val].transform(_ecdf_y, complementary)

    cols += [col]

    return _cat_source(df, cats, cols, None)


def _ecdf_collection_dots(
    df, val, cats, cols, complementary, order, palette, show_legend, y, p, **kwargs
):
    _, _, color_factors = _get_cat_range(df, df.groupby(cats), order, None, False)

    source = _point_ecdf_source(df, val, cats, cols, complementary, False)

    if "color" not in kwargs:
        kwargs["color"] = bokeh.transform.factor_cmap(
            "cat", palette=palette, factors=color_factors
        )

    if show_legend:
        kwargs["legend"] = "__label"

    p.circle(source=source, x=val, y=y, **kwargs)

    return p


def _ecdf_collection_staircase(
    df, val, cats, complementary, order, palette, show_legend, p, **kwargs
):
    grouped = df.groupby(cats)

    color_not_in_kwargs = "color" not in kwargs

    if order is None:
        order = list(grouped.groups.keys())
    grouped_iterator = [
        (order_val, grouped.get_group(order_val)) for order_val in order
    ]

    for i, g in enumerate(grouped_iterator):
        if show_legend:
            if type(g[0]) == tuple:
                legend = ", ".join([str(c) for c in g[0]])
            else:
                legend = str(g[0])
        else:
            legend = None

        if color_not_in_kwargs:
            kwargs["color"] = palette[i % len(palette)]

        _ecdf(
            g[1][val],
            staircase=True,
            p=p,
            legend=legend,
            complementary=complementary,
            **kwargs,
        )

    return p


def bokeh_imrgb(im, plot_height=400, plot_width=None,
                tools='pan,box_zoom,wheel_zoom,reset,resize'):
    """
    Make a Bokeh Figure instance displaying an RGB image.
    If the image is already 32 bit, just display it
    """
    # Make 32 bit image
    if len(im.shape) == 2 and im.dtype == np.uint32:
        im_disp = im
    else:
        im_disp = rgb_to_rgba32(im)

    # Get shape
    n, m = im_disp.shape

    # Determine plot height and width
    if plot_height is not None and plot_width is None:
        plot_width = int(m/n * plot_height)
    elif plot_height is None and plot_width is not None:
        plot_height = int(n/m * plot_width)
    elif plot_height is None and plot_width is None:
        plot_heigt = 400
        plot_width = int(m/n * plot_height)

    # Set up figure with appropriate dimensions
    p = bokeh.plotting.figure(plot_height=plot_height, plot_width=plot_width,
                              x_range=[0, m], y_range=[0, n], tools=tools)

    # Display the image, setting the origin and heights/widths properly
    p.image_rgba(image=[im_disp], x=0, y=0, dw=m, dh=n)

    return p


def bokeh_im(im, plot_height=400, plot_width=None,
             color_palette=bokeh.palettes.gray(256),
             tools='pan,box_zoom,wheel_zoom,reset,resize'):
    """
    """
    # Get shape
    n, m = im.shape

    # Determine plot height and width
    if plot_height is not None and plot_width is None:
        plot_width = int(m/n * plot_height)
    elif plot_height is None and plot_width is not None:
        plot_height = int(n/m * plot_width)
    elif plot_height is None and plot_width is None:
        plot_heigt = 400
        plot_width = int(m/n * plot_height)

    p = bokeh.plotting.figure(plot_height=plot_height, plot_width=plot_width,
                              x_range=[0, m], y_range=[0, n], tools=tools)

    # Set color mapper
    color = bokeh.models.LinearColorMapper(color_palette)

    # Display the image
    p.image(image=[im], x=0, y=0, dw=m, dh=n, color_mapper=color)

    return p


def distribution_plot_app(
    x_min=None,
    x_max=None,
    scipy_dist=None,
    transform=None,
    custom_pdf=None,
    custom_pmf=None,
    custom_cdf=None,
    params=None,
    n=400,
    plot_height=200,
    plot_width=300,
    x_axis_label="x",
    title=None,
):
    """
    Build interactive Bokeh app displaying a univariate
    probability distribution.

    Parameters
    ----------
    x_min : float
        Minimum value that the random variable can take in plots.
    x_max : float
        Maximum value that the random variable can take in plots.
    scipy_dist : scipy.stats distribution
        Distribution to use in plotting.
    transform : function or None (default)
        A function of call signature `transform(*params)` that takes
        a tuple or Numpy array of parameters and returns a tuple of
        the same length with transformed parameters.
    custom_pdf : function
        Function with call signature f(x, *params) that computes the
        PDF of a distribution.
    custom_pmf : function
        Function with call signature f(x, *params) that computes the
        PDF of a distribution.
    custom_cdf : function
        Function with call signature F(x, *params) that computes the
        CDF of a distribution.
    params : list of dicts
        A list of parameter specifications. Each entry in the list gives
        specifications for a parameter of the distribution stored as a
        dictionary. Each dictionary must have the following keys.
            name : str, name of the parameter
            start : float, starting point of slider for parameter (the
                smallest allowed value of the parameter)
            end : float, ending point of slider for parameter (the
                largest allowed value of the parameter)
            value : float, the value of the parameter that the slider
                takes initially. Must be between start and end.
            step : float, the step size for the slider
    n : int, default 400
        Number of points to use in making plots of PDF and CDF for
        continuous distributions. This should be large enough to give
        smooth plots.
    plot_height : int, default 200
        Height of plots.
    plot_width : int, default 300
        Width of plots.
    x_axis_label : str, default 'x'
        Label for x-axis.
    title : str, default None
        Title to be displayed above the PDF or PMF plot.

    Returns
    -------
    output : Bokeh app
        An app to visualize the PDF/PMF and CDF. It can be displayed
        with bokeh.io.show(). If it is displayed in a notebook, the
        notebook_url kwarg should be specified.
    """
    if None in [x_min, x_max]:
        raise RuntimeError("`x_min` and `x_max` must be specified.")

    if scipy_dist is None:
        fun_c = custom_cdf
        if (custom_pdf is None and custom_pmf is None) or custom_cdf is None:
            raise RuntimeError(
                "For custom distributions, both PDF/PMF and" + " CDF must be specified."
            )
        if custom_pdf is not None and custom_pmf is not None:
            raise RuntimeError("Can only specify custom PMF or PDF.")
        if custom_pmf is None:
            discrete = False
            fun_p = custom_pdf
        else:
            discrete = True
            fun_p = custom_pmf
    elif custom_pdf is not None or custom_pmf is not None or custom_cdf is not None:
        raise RuntimeError("Can only specify either custom or scipy distribution.")
    else:
        fun_c = scipy_dist.cdf
        if hasattr(scipy_dist, "pmf"):
            discrete = True
            fun_p = scipy_dist.pmf
        else:
            discrete = False
            fun_p = scipy_dist.pdf

    if discrete:
        p_y_axis_label = "PMF"
    else:
        p_y_axis_label = "PDF"

    if params is None:
        raise RuntimeError("`params` must be specified.")

    def _plot_app(doc):
        p_p = bokeh.plotting.figure(
            plot_height=plot_height,
            plot_width=plot_width,
            x_axis_label=x_axis_label,
            y_axis_label=p_y_axis_label,
            title=title,
        )
        p_c = bokeh.plotting.figure(
            plot_height=plot_height,
            plot_width=plot_width,
            x_axis_label=x_axis_label,
            y_axis_label="CDF",
        )

        # Link the axes
        p_c.x_range = p_p.x_range

        # Make sure CDF y_range is zero to one
        p_c.y_range = bokeh.models.Range1d(-0.05, 1.05)

        # Make array of parameter values
        param_vals = np.array([param["value"] for param in params])
        if transform is not None:
            param_vals = transform(*param_vals)

        # Set up data for plot
        if discrete:
            x = np.arange(int(np.ceil(x_min)), int(np.floor(x_max)) + 1)
            x_size = x[-1] - x[0]
            x_c = np.empty(2 * len(x))
            x_c[::2] = x
            x_c[1::2] = x
            x_c = np.concatenate(
                (
                    (max(x[0] - 0.05 * x_size, x[0] - 0.95),),
                    x_c,
                    (min(x[-1] + 0.05 * x_size, x[-1] + 0.95),),
                )
            )
            x_cdf = np.concatenate(((x_c[0],), x))
        else:
            x = np.linspace(x_min, x_max, n)
            x_c = x_cdf = x

        # Compute PDF and CDF
        y_p = fun_p(x, *param_vals)
        y_c = fun_c(x_cdf, *param_vals)
        if discrete:
            y_c_plot = np.empty_like(x_c)
            y_c_plot[::2] = y_c
            y_c_plot[1::2] = y_c
            y_c = y_c_plot

        # Set up data sources
        source_p = bokeh.models.ColumnDataSource(data={"x": x, "y_p": y_p})
        source_c = bokeh.models.ColumnDataSource(data={"x": x_c, "y_c": y_c})

        # Plot PDF and CDF
        p_c.line("x", "y_c", source=source_c, line_width=2)
        if discrete:
            p_p.circle("x", "y_p", source=source_p, size=5)
            p_p.segment(x0="x", x1="x", y0=0, y1="y_p", source=source_p, line_width=2)
        else:
            p_p.line("x", "y_p", source=source_p, line_width=2)

        def _callback(attr, old, new):
            param_vals = tuple([slider.value for slider in sliders])
            if transform is not None:
                param_vals = transform(*param_vals)

            # Compute PDF and CDF
            source_p.data["y_p"] = fun_p(x, *param_vals)
            y_c = fun_c(x_cdf, *param_vals)
            if discrete:
                y_c_plot = np.empty_like(x_c)
                y_c_plot[::2] = y_c
                y_c_plot[1::2] = y_c
                y_c = y_c_plot
            source_c.data["y_c"] = y_c

        sliders = [
            bokeh.models.Slider(
                start=param["start"],
                end=param["end"],
                value=param["value"],
                step=param["step"],
                title=param["name"],
            )
            for param in params
        ]
        for slider in sliders:
            slider.on_change("value", _callback)

        # Add the plot to the app
        widgets = bokeh.layouts.widgetbox(sliders)
        grid = bokeh.layouts.gridplot([p_p, p_c], ncols=2)
        doc.add_root(bokeh.layouts.column(widgets, grid))

    handler = bokeh.application.handlers.FunctionHandler(_plot_app)
    return bokeh.application.Application(handler)


def adjust_range(element, buffer=0.05):
    """
    Adjust soft ranges of dimensions of HoloViews element.

    Parameters
    ----------
    element : holoviews element
        Element which will have the `soft_range` of each kdim and vdim
        recomputed to give a buffer around the glyphs.
    buffer : float, default 0.05
        Buffer, as a fraction of the whole data range, to give around
        data.

    Returns
    -------
    output : holoviews element
        Inputted HoloViews element with updated soft_ranges for its
        dimensions.
    """
    # This only works with DataFrames
    if type(element.data) != pd.core.frame.DataFrame:
        raise RuntimeError("Can only adjust range if data is Pandas DataFrame.")

    # Adjust ranges of kdims
    for i, dim in enumerate(element.kdims):
        if element.data[dim.name].dtype in [float, int]:
            data_range = (element.data[dim.name].min(), element.data[dim.name].max())
            if data_range[1] - data_range[0] > 0:
                buff = buffer * (data_range[1] - data_range[0])
                element.kdims[i].soft_range = (
                    data_range[0] - buff,
                    data_range[1] + buff,
                )

    # Adjust ranges of vdims
    for i, dim in enumerate(element.vdims):
        if element.data[dim.name].dtype in [float, int]:
            data_range = (element.data[dim.name].min(), element.data[dim.name].max())
            if data_range[1] - data_range[0] > 0:
                buff = buffer * (data_range[1] - data_range[0])
                element.vdims[i].soft_range = (
                    data_range[0] - buff,
                    data_range[1] + buff,
                )

    return element




try:
    import multiprocess
except:
    import multiprocessing as multiprocess

import warnings

import tqdm

import numpy as np
import numba

from . import utils


def seed_rng(seed):
    """
    Seed random number generators for Numpy and Numba'd functions.

    Parameters
    ----------
    seed : long int
        Seed for RNG
    """
    try:

        @numba.jit(nopython=True)
        def _seed(seed):
            np.random.seed(seed)

        _seed(seed)
    except:
        warnings.warn(
            "Unable to seed Numba RNG. It is possible Numba is not"
            " properly installed. If that is the case, you all bootstrap calculations"
            " will use un-Numba'd random number generation that is properly seeded."
            " However, if Numba is installed and there is some other issue preventing"
            " proper seeding of the random number generator used in Numba'd code,"
            " you may get unexpected results."
        )

    np.random.seed(seed)


def draw_bs_reps(data, func, size=1, args=()):
    """
    Generate bootstrap replicates out of `data` using `func`.

    Parameters
    ----------
    data : array_like
        One-dimensional array of data.
    func : function
        Function, with call signature `func(data, *args)` to compute
        replicate statistic from resampled `data`.
    size : int, default 1
        Number of bootstrap replicates to generate.
    args : tuple, default ()
        Arguments to be passed to `func`.

    Returns
    -------
    output : ndarray
        Bootstrap replicates computed from `data` using `func`.

    Notes
    -----
    nan values are ignored.
    """
    data = utils._convert_data(data)

    if args == ():
        if func == np.mean:
            return _draw_bs_reps_mean(data, size=size)
        elif func == np.median:
            return _draw_bs_reps_median(data, size=size)
        elif func == np.std:
            return _draw_bs_reps_std(data, size=size)

    # Make Numba'd function
    f, numba_success = utils._make_one_arg_numba_func(func, args)

    if numba_success:
        jit = numba.jit
    else:
        jit = utils._dummy_jit

    @jit(nopython=True)
    def _draw_bs_reps(data):
        # Set up output array
        bs_reps = np.empty(size)

        # Draw replicates
        n = len(data)
        for i in range(size):
            bs_reps[i] = f(np.random.choice(data, size=n), args)

        return bs_reps

    return _draw_bs_reps(data)


@numba.jit(nopython=True)
def _draw_bs_reps_mean(data, size=1):
    """
    Generate bootstrap replicates of the mean out of `data`.

    Parameters
    ----------
    data : array_like
        One-dimensional array of data.
    size : int, default 1
        Number of bootstrap replicates to generate.

    Returns
    -------
    output : float
        Bootstrap replicates of the mean computed from `data`.
    """
    # Set up output array
    bs_reps = np.empty(size)

    # Draw replicates
    n = len(data)
    for i in range(size):
        bs_reps[i] = np.mean(np.random.choice(data, size=n))

    return bs_reps


@numba.jit(nopython=True)
def _draw_bs_reps_median(data, size=1):
    """
    Generate bootstrap replicates of the median out of `data`.

    Parameters
    ----------
    data : array_like
        One-dimensional array of data.
    size : int, default 1
        Number of bootstrap replicates to generate.

    Returns
    -------
    output : float
        Bootstrap replicates of the median computed from `data`.
    """
    # Set up output array
    bs_reps = np.empty(size)

    # Draw replicates
    n = len(data)
    for i in range(size):
        bs_reps[i] = np.median(np.random.choice(data, size=n))

    return bs_reps


@numba.jit(nopython=True)
def _draw_bs_reps_std(data, ddof=0, size=1):
    """
    Generate bootstrap replicates of the median out of `data`.

    Parameters
    ----------
    data : array_like
        One-dimensional array of data.
    ddof : int
        Delta degrees of freedom. Divisor in standard deviation
        calculation is `len(data) - ddof`.
    size : int, default 1
        Number of bootstrap replicates to generate.

    Returns
    -------
    output : float
        Bootstrap replicates of the median computed from `data`.
    """
    # Set up output array
    bs_reps = np.empty(size)

    # Draw replicates
    n = len(data)
    for i in range(size):
        bs_reps[i] = np.std(np.random.choice(data, size=n))

    if ddof > 0:
        return bs_reps * np.sqrt(n / (n - ddof))

    return bs_reps


def draw_bs_reps_pairs(x, y, func, size=1, args=()):
    """
    Perform pairs bootstrap for single statistic.

    Parameters
    ----------
    x : array_like
        x-values of data.
    y : array_like
        y-values of data.
    func : function
        Function, with call signature `func(x, y, *args)` to compute
        replicate statistic from pairs bootstrap sample. It must return
        a single, scalar value.
    size : int, default 1
        Number of pairs bootstrap replicates to draw.
    args : tuple, default ()
        Arguments to be passed to `func`.

    Returns
    -------
    output : ndarray
        Bootstrap replicates.
    """
    x, y = utils._convert_two_data(x, y, min_len=1)

    # Make Numba'd function
    f, numba_success = utils._make_two_arg_numba_func(func, args)

    if numba_success:
        jit = numba.jit
    else:
        jit = utils._dummy_jit

    n = len(x)

    @jit(nopython=True)
    def _draw_bs_reps_pairs(x, y):
        # Set up array of indices to sample from
        inds = np.arange(n)

        # Initialize replicates
        bs_replicates = np.empty(size)

        # Generate replicates
        for i in range(size):
            bs_inds = np.random.choice(inds, n)
            bs_x, bs_y = x[bs_inds], y[bs_inds]
            bs_replicates[i] = f(bs_x, bs_y, args)

        return bs_replicates

    return _draw_bs_reps_pairs(x, y)


def draw_perm_reps(data_1, data_2, func, size=1, args=()):
    """
    Generate permutation replicates of `func` from `data_1` and
    `data_2`

    Parameters
    ----------
    data_1 : array_like
        One-dimensional array of data.
    data_2 : array_like
        One-dimensional array of data.
    func : function
        Function, with call signature `func(x, y, *args)` to compute
        replicate statistic from permutation sample. It must return
        a single, scalar value.
    size : int, default 1
        Number of pairs bootstrap replicates to draw.
    args : tuple, default ()
        Arguments to be passed to `func`.

    Returns
    -------
    output : ndarray
        Permutation replicates.
    """
    # Convert to Numpy arrays
    data_1 = utils._convert_data(data_1)
    data_2 = utils._convert_data(data_2)

    if args == ():
        if func == diff_of_means:
            return _draw_perm_reps_diff_of_means(data_1, data_2, size=size)
        elif func == studentized_diff_of_means:
            if len(data_1) == 1 or len(data_2) == 1:
                raise RuntimeError("Data sets must have at least two entries")
            return _draw_perm_reps_studentized_diff_of_means(data_1, data_2, size=size)

    # Make a Numba'd function for drawing reps.
    f, numba_success = utils._make_two_arg_numba_func(func, args)

    if numba_success:
        jit = numba.jit
    else:
        jit = utils._dummy_jit

    @jit(nopython=True)
    def _draw_perm_reps(data_1, data_2):
        n1 = len(data_1)
        x = np.concatenate((data_1, data_2))

        perm_reps = np.empty(size)
        for i in range(size):
            np.random.shuffle(x)
            perm_reps[i] = f(x[:n1], x[n1:], args)

        return perm_reps

    return _draw_perm_reps(data_1, data_2)


@numba.jit(nopython=True)
def _draw_perm_reps_diff_of_means(data_1, data_2, size=1):
    """
    Generate permutation replicates of difference of means from
    `data_1` and `data_2`

    Parameters
    ----------
    data_1 : array_like
        One-dimensional array of data.
    data_2 : array_like
        One-dimensional array of data.
    size : int, default 1
        Number of pairs bootstrap replicates to draw.

    Returns
    -------
    output : ndarray
        Permutation replicates.
    """
    n1 = len(data_1)
    x = np.concatenate((data_1, data_2))

    perm_reps = np.empty(size)
    for i in range(size):
        np.random.shuffle(x)
        perm_reps[i] = _diff_of_means(x[:n1], x[n1:])

    return perm_reps


@numba.jit(nopython=True)
def _draw_perm_reps_studentized_diff_of_means(data_1, data_2, size=1):
    """
    Generate permutation replicates of Studentized difference
    of means from  `data_1` and `data_2`

    Parameters
    ----------
    data_1 : array_like
        One-dimensional array of data.
    data_2 : array_like
        One-dimensional array of data.
    size : int, default 1
        Number of pairs bootstrap replicates to draw.

    Returns
    -------
    output : ndarray
        Permutation replicates.
    """
    n1 = len(data_1)
    x = np.concatenate((data_1, data_2))

    perm_reps = np.empty(size)
    for i in range(size):
        np.random.shuffle(x)
        perm_reps[i] = _studentized_diff_of_means(x[:n1], x[n1:])

    return perm_reps


def diff_of_means(data_1, data_2):
    """
    Difference in means of two arrays.

    Parameters
    ----------
    data_1 : array_like
        One-dimensional array of data.
    data_2 : array_like
        One-dimensional array of data.

    Returns
    -------
    output : float
        np.mean(data_1) - np.mean(data_2)
    """
    data_1 = utils._convert_data(data_1)
    data_2 = utils._convert_data(data_2)

    return _diff_of_means(data_1, data_2)


@numba.jit(nopython=True)
def _diff_of_means(data_1, data_2):
    """
    Difference in means of two arrays.

    Parameters
    ----------
    data_1 : array_like
        One-dimensional array of data.
    data_2 : array_like
        One-dimensional array of data.

    Returns
    -------
    output : float
        np.mean(data_1) - np.mean(data_2)
    """
    return np.mean(data_1) - np.mean(data_2)


def studentized_diff_of_means(data_1, data_2):
    """
    Studentized difference in means of two arrays.

    Parameters
    ----------
    data_1 : array_like
        One-dimensional array of data.
    data_2 : array_like
        One-dimensional array of data.

    Returns
    -------
    output : float
        Studentized difference of means.

    Notes
    -----
    If the variance of both `data_1` and `data_2` is zero, returns
    np.nan.
    """
    data_1 = utils._convert_data(data_1, min_len=2)
    data_2 = utils._convert_data(data_2, min_len=2)

    return _studentized_diff_of_means(data_1, data_2)


@numba.jit(nopython=True)
def _studentized_diff_of_means(data_1, data_2):
    """
    Studentized difference in means of two arrays.

    Parameters
    ----------
    data_1 : array_like
        One-dimensional array of data.
    data_2 : array_like
        One-dimensional array of data.

    Returns
    -------
    output : float
        Studentized difference of means.

    Notes
    -----
    If the variance of both `data_1` and `data_2` is zero, returns
    np.nan.
    """
    if utils._allequal(data_1) and utils._allequal(data_2):
        return np.nan

    denom = np.sqrt(
        np.var(data_1) / (len(data_1) - 1) + np.var(data_2) / (len(data_2) - 1)
    )

    return (np.mean(data_1) - np.mean(data_2)) / denom


def pearson_r(data_1, data_2):
    """
    Compute the Pearson correlation coefficient between two samples.

    Parameters
    ----------
    data_1 : array_like
        One-dimensional array of data.
    data_2 : array_like
        One-dimensional array of data.

    Returns
    -------
    output : float
        The Pearson correlation coefficient between `data_1`
        and `data_2`.

    Notes
    -----
    Only entries where both `data_1` and `data_2` are not NaN are
    used.

    If the variance of `data_1` or `data_2` is zero, return NaN.
    """
    x, y = utils._convert_two_data(data_1, data_2, inf_ok=False, min_len=2)
    return utils._pearson_r(x, y)


def _draw_bs_reps_mle(
    mle_fun,
    gen_fun,
    data,
    mle_args=(),
    gen_args=(),
    size=1,
    progress_bar=False,
    rng=None,
):
    """Draw parametric bootstrap replicates of maximum likelihood
    estimator.

    Parameters
    ----------
    mle_fun : function
        Function with call signature `mle_fun(data, *mle_args)` that
        computes a MLE for the parameters
    gen_fun : function
        Function to randomly draw a new data set out of the model
        distribution parametrized by the MLE. Must have call
        signature `gen_fun(params, *gen_args, size, rng)`. Note
        that `size` in as an argument in this function relates to the
        number of data you will generate, which is always equal to
        len(data). This is not the same as the `size` argument of
        `_draw_bs_reps_mle()`, which is the number of bootstrap
        replicates you wish to draw.
    data : Numpy array, possibly multidimensional
        Array of measurements. The first index should index repeat of
        experiment. E.g., if the data consist of n (x, y) pairs, `data`
        should have shape (n, 2).
    mle_args : tuple, default ()
        Arguments to be passed to `mle_fun()`.
    gen_args : tuple, default ()
        Arguments to be passed to `gen_fun()`.
    size : int, default 1
        Number of bootstrap replicates to draw.
    progress_bar : bool, default False
        Whether or not to display progress bar.
    rng : numpy.random.Generator instance, default None
        RNG to be used in bootstrapping. If None, the default
        Numpy RNG is used with a fresh seed based on the clock.

    Returns
    -------
    output : numpy array
        Bootstrap replicates of MLEs.
    """
    if rng is None:
        rng = np.random.default_rng()

    params = mle_fun(data, *mle_args)

    if progress_bar:
        iterator = tqdm.tqdm(range(size))
    else:
        iterator = range(size)

    return np.array(
        [
            mle_fun(gen_fun(params, *gen_args, size=len(data), rng=rng), *mle_args)
            for _ in iterator
        ]
    )


def draw_bs_reps_mle(
    mle_fun,
    gen_fun,
    data,
    mle_args=(),
    gen_args=(),
    size=1,
    n_jobs=1,
    progress_bar=False,
    rng=None,
):
    """Draw bootstrap replicates of maximum likelihood estimator.

    Parameters
    ----------
    mle_fun : function
        Function with call signature `mle_fun(data, *mle_args)` that
        computes a MLE for the parameters.
    gen_fun : function
        Function to randomly draw a new data set out of the model
        distribution parametrized by the MLE. Must have call
        signature `gen_fun(params, *gen_args, size, rng)`. Note
        that `size` as an argument in this function relates to the
        number of data you will generate, which is always equal to
        len(data). This is not the same as the `size` argument of
        `draw_bs_reps_mle()`, which is the number of bootstrap
        replicates you wish to draw.
    data : Numpy array, possibly multidimensional
        Array of measurements. The first index should index repeat of
        experiment. E.g., if the data consist of n (x, y) pairs, `data`
        should have shape (n, 2).
    mle_args : tuple, default ()
        Arguments to be passed to `mle_fun()`.
    gen_args : tuple, default ()
        Arguments to be passed to `gen_fun()`.
    size : int, default 1
        Number of bootstrap replicates to draw.
    n_jobs : int, default 1
        Number of cores to use in drawing bootstrap replicates.
    progress_bar : bool, default False
        Whether or not to display progress bar.
    rng : numpy.random.Generator instance, default None
        RNG to be used in bootstrapping. If None, the default
        Numpy RNG is used with a fresh seed based on the clock.

    Returns
    -------
    output : numpy array
        Bootstrap replicates of MLEs.
    """
    # Just call the original function if n_jobs is 1 (no parallelization)
    if n_jobs == 1:
        return _draw_bs_reps_mle(
            mle_fun,
            gen_fun,
            data,
            mle_args=mle_args,
            gen_args=gen_args,
            size=size,
            progress_bar=progress_bar,
            rng=rng,
        )

    if rng is not None:
        raise RuntimeError(
            "You are attempting to draw replicates in parallel with a specified random"
            " number generator (`rng` is not `None`). Each of the sets of replicates"
            " drawn in parallel will be the same since the random number generator is"
            " not reseeded for each thread. When running in parallel, you  must have"
            " `rng=None`."
        )

    # Set up sizes of bootstrap replicates for each core, making sure we
    # get all of them, even if sizes % n_jobs != 0
    sizes = [size // n_jobs for _ in range(n_jobs)]
    sizes[-1] += size - sum(sizes)

    # Build arguments
    arg_iterable = [
        (mle_fun, gen_fun, data, mle_args, gen_args, s, progress_bar, None)
        for s in sizes
    ]

    with multiprocess.Pool(n_jobs) as pool:
        result = pool.starmap(_draw_bs_reps_mle, arg_iterable)

    return np.concatenate(result)



"""Command-line interface for running SBC analysis."""
import pathlib
import tomllib

import click
import cmdstanpy

from . import stan


def parse_toml_config(toml_path):
    """Parse TOML configuration file and extract SBC parameters.

    Parameters
    ----------
    toml_path : str or pathlib.Path
        Path to the TOML configuration file.

    Returns
    -------
    config : dict
        Dictionary containing all SBC configuration parameters.
    """
    toml_path = pathlib.Path(toml_path)

    with open(toml_path, "rb") as f:
        config = tomllib.load(f)

    return config


@click.command()
@click.argument("config_file", type=click.Path(exists=True))
def main(config_file):
    """Run SBC analysis from a TOML configuration file.

    All parameters including output file path are specified in the config file.

    CONFIG_FILE: Path to TOML configuration file
    """
    click.echo(f"Reading configuration from: {config_file}")
    config = parse_toml_config(config_file)

    # Get required fields from TOML file
    required_keys = (
        'prior_predictive_stan_file',
        'posterior_stan_file',
        'prior_predictive_model_data',
        'posterior_model_data',
        'measured_data'
    )
    sbc_kwargs = {}
    for key in required_keys:
        if key in config:
            sbc_kwargs[key] = config[key]
        else:
            raise RuntimeError(f'{key} must be specified.')

    # Get optional fields
    sbc_kwargs['var_names'] = config.get('var_names', None)
    sbc_kwargs['measured_data_dtypes'] = config.get('measured_data_dtypes', {})
    sbc_kwargs['posterior_predictive_var_names'] = config.get('posterior_predictive_var_names', None)
    sbc_kwargs['log_likelihood_var_name'] = config.get('log_likelihood_var_name', None)
    sbc_kwargs['sampling_kwargs'] = config.get('sampling_kwargs', {})
    sbc_kwargs['diagnostic_check_kwargs'] = config.get('diagnostic_check_kwargs', {})
    sbc_kwargs['cores'] = config.get('cores', 1)
    sbc_kwargs['N'] = config.get('N', 400)
    sbc_kwargs['n_prior_draws_for_sd'] = config.get('n_prior_draws_for_sd', 1000)
    sbc_kwargs['samples_dir'] = config.get('samples_dir', 'sbc_samples')
    sbc_kwargs['remove_sample_files'] = config.get('remove_sample_files', True)
    sbc_kwargs['progress_bar'] = config.get('progress_bar', False)
    sbc_kwargs['output_csv'] = config.get('output_csv', 'sbc_results.csv')
    sbc_kwargs['overwrite'] = config.get('overwrite', False)

    # Run SBC (results written directly to CSV as workers complete)
    click.echo("Running SBC analysis...")
    click.echo(f"Results will be written to: {sbc_kwargs['output_csv']}\n")

    result_df = stan.sbc(**sbc_kwargs)

    click.echo(f"\nSBC analysis complete! Results saved to: {sbc_kwargs['output_csv']}")
    click.echo(f"Total rows: {len(result_df)}")


if __name__ == "__main__":
    main()



import re
import datetime
import time
import random
import contextlib
import copy
import glob
import logging
import os
import shutil
import sys
import tomllib
import urllib
import warnings

import multiprocessing

import tqdm

import numpy as np
import pandas as pd
import polars as pl
import xarray

try:
    import arviz as az
    import arviz.sel_utils
except:
    raise RuntimeError("Could not import ArviZ. Perhaps it is not installed.")

try:
    import cmdstanpy
    cmdstanpy_success = True
except:
    cmdstanpy_success = False

if not cmdstanpy_success:
    raise RuntimeError("CmdStanPy could not be imported.")


def StanModel(
    file=None,
    charset="utf-8",
    model_name="anon_model",
    model_code=None,
    force_compile=False,
    **kwargs,
):
    """ "Utility to load/save cached compiled Stan models using PyStan.
    DEPRECATED
    """
    raise RuntimeError('bebi103.stan.StanModel is deprecated.')


def install_cmdstan_colab(cmdstan_colab_url=None, quiet=False):
    """Install CmdStan with Google Colab.

    Parameters
    ----------
    cmdstan_colab_url : str, default None
        Full URL for zip file containing version of Stan for
        installation. If None, searches for the most recent version. In
        general, this should be `None`.
    quiet : bool, default False
        If True, suppresses success message.

    Returns
    -------
    None
    """
    if "google.colab" not in sys.modules:
        raise RuntimeError('This is only meant for installation using Google Colab.')

    if cmdstan_colab_url is None:
        from cmdstanpy.install_cmdstan import latest_version
        cmdstan_version = latest_version()

        # First spelling of URL
        try:
            cmdstan_url = f"https://github.com/stan-dev/cmdstan/releases/download/v{cmdstan_version}/"
            fname = f"colab-cmdstan-{cmdstan_version}.tgz"
            urllib.request.urlretrieve(cmdstan_url + fname, fname)
            shutil.unpack_archive(fname)
            os.environ["CMDSTAN"] = f"./cmdstan-{cmdstan_version}"

            if not quiet:
                print(f'CmdStan version {cmdstan_version} successfully installed.')

            return None
        except:
            pass

        # Old misspelling
        try:
            cmdstan_url = f"https://github.com/stan-dev/cmdstan/releases/download/v{cmdstan_version}/"
            fname = f"collab-cmdstan-{cmdstan_version}.tgz"
            urllib.request.urlretrieve(cmdstan_url + fname, fname)
            shutil.unpack_archive(fname)
            os.environ["CMDSTAN"] = f"./cmdstan-{cmdstan_version}"

            if not quiet:
                print(f'CmdStan version {cmdstan_version} successfully installed.')

            return None
        except:
            pass

        # Try each with .tar.gz suffix
        try:
            cmdstan_url = f"https://github.com/stan-dev/cmdstan/releases/download/v{cmdstan_version}/"
            fname = f"colab-cmdstan-{cmdstan_version}.tar.gz"
            urllib.request.urlretrieve(cmdstan_url + fname, fname)
            shutil.unpack_archive(fname)
            os.environ["CMDSTAN"] = f"./cmdstan-{cmdstan_version}"

            if not quiet:
                print(f'CmdStan version {cmdstan_version} successfully installed.')

            return None
        except:
            pass

        try:
            cmdstan_url = f"https://github.com/stan-dev/cmdstan/releases/download/v{cmdstan_version}/"
            fname = f"collab-cmdstan-{cmdstan_version}.tar.gz"
            urllib.request.urlretrieve(cmdstan_url + fname, fname)
            shutil.unpack_archive(fname)
            os.environ["CMDSTAN"] = f"./cmdstan-{cmdstan_version}"

            if not quiet:
                print(f'CmdStan version {cmdstan_version} successfully installed.')

            return None
        except:
            pass

        # The only way we get here is if we never successfully completed a try
        raise RuntimeError('Unable to install CmdStan, most likely because the URL for the most recent version of CmdStan could not be found. Try finding at URL for a downloadable Colab-compatible version of CmdStan here: https://github.com/stan-dev/cmdstan/releases. Then call `install_cmdstan_colab(cmdstan_colab_url)`.')
    else:
        try:
            urllib.request.urlretrieve(cmdstan_colab_url)
            shutil.unpack_archive(fname)
            os.environ["CMDSTAN"] = f"./cmdstan-{cmdstan_version}"

            if not quiet:
                print('CmdStan successfully installed. You should nonetheless test the installation.')

            return None
        except:
            raise RuntimeError('Unable to install CmdStan, most likely because the inputted URL for CmdStan was invalid. You can run without providing a URL, and the latest version of CmdStan willbe installed; `install_cmdstan_colab()`.')


def include_path():
    """Return path to include files for Stan.

    Returns
    -------
    output : str
        Absolute path to directory containing include files.
    """
    return os.path.join(os.path.dirname(__file__), 'stan_include')


def clean_cmdstan(path="./", prefix=None, delete_sampling_output=False):
    """Remove all .hpp, .o, .d, and executable files resulting from
    compilation of Stan models using CmdStanPy.

    Parameters
    ----------
    path : str, default './'
        Path to directory containing files to delete.
    prefix : str, default None
        String of prefix of model name. This is the name of the Stan
        file from which the model was generated is <prefix>.stan. If
        None, then all .stan files are used. Files <prefix>.hpp,
        <prefix>.o, <prefix>.d, and the executable file <prefix> are
        deleted.
    delete_sampling_output: bool, default False
        If True, also delete all output generated by CmdStan.

    Notes
    -----
    If your files are stored in a temporary directory, as for the
    CmdStanPy default, use `cmdstanpy.cleanup_tmpdir()` instead.



    """
    if path == "/":
        raise RuntimeError(
            "No way you can delete stuff from root. You are making a big mistake."
        )

    if prefix is None:
        prefix = ""

    stan_files = glob.glob(os.path.join(path, prefix, "*.stan"))
    prefixes = [fname[:-5] for fname in stan_files]

    hpp_files = [prefix + ".hpp" for prefix in prefixes]
    o_files = [prefix + ".o" for prefix in prefixes]
    d_files = [prefix + ".d" for prefix in prefixes]

    for fname in hpp_files + o_files + d_files:
        if os.path.isfile(fname):
            os.remove(fname)

    for fname in prefixes:
        if os.path.isfile(fname) and os.access(fname, os.X_OK):
            os.remove(fname)

    if delete_sampling_output:
        for pre in prefixes:
            out_files = glob.glob(pre + "*.csv")
            out_files += glob.glob(pre + "*.txt")
            for fname in out_files:
                if os.path.isfile(fname):
                    os.remove(fname)


def cmdstan_version():
    """Determine CmdStan version

    Returns
    -------
    output : str
        Version of installed CmdStan used by CmdStanPy. If unable to
        ascertain version, return a string:
        "Unable to determine CmdStan version."

    Notes
    -----
    Only works if CmdStanPy is installed.
    """
    try:
        cmdstan_path = cmdstanpy.cmdstan_path()
        return cmdstan_path[cmdstan_path.rfind("-") + 1 :]
    except:
        return "Unable to determine CmdStan version."


def df_to_datadict_hier(
    df=None, level_cols=None, data_cols=None, sort_cols=[], cowardly=False
):
    """Convert a tidy data frame to a data dictionary for a hierarchical
    Stan model.

    Parameters
    ----------
    df : DataFrame
        A tidy Polars or Pandas data frame.
    level_cols : list
        A list of column names containing variables that specify the
        level of the hierarchical model. These must be given in order
        of the hierarchy of levels, with the first entry being the
        farthest from the data.
    data_cols : list
        A list of column names containing the data.
    sort_cols : list, default []]
        List of columns names to use in sorting the data. They will be
        first sorted by the level indices, and the subsequently sorted
        according to sort_cols.
    cowardly : bool, default False
        If True, refuse to generate new columns if they already exist
        in the data frame. If you run this function using a data frame
        that was outputted previously by this function, you will get an
        error if `cowardly` is True. Otherwise, the columns may be
        overwritten.

    Returns
    -------
    data : dict
        A dictionary that can be passed to into a Stan program. The
        dictionary contains keys/entries:

            - `'N'`: Total number of data points
            - `'J_1'`: Number of hyper parameters for hierarchical level 1.
            - `'J_2'`: Number of hyper parameters for hierarchical level 2. ... and so on with `'J_3'`, `'J_4'`, ...
            - `'index_1'`: Set of `J_2` indices defining which level 1 parameters condition the level 2 parameters.
            - `'index_2'`: Set of `J_3` indices defining which level 2 parameters condition the level 3 parameters. ...and so on for 'index_3', etc.
            - `'index_k'`: Set of `N` indices defining which of the level k parameters condition the data, for a k-level hierarchical model.
            - `data_col[0]` : Data from first data_col
            - `data_col[1]` : Data from second data_col ...and so on.

    df : DataFrame
        Updated input data frame with added columnes with names given by
        `level_col[0] + '_stan'`, `level_col[1] + '_stan'`, etc. These
        contain the integer indices that correspond to the possibly
        non-integer values in the `level_col`s of the original data
        frame. This enables interpretation of Stan results, which have
        everything integer indexed.

    Notes
    -----
    Assumes no missing data.

    The ordering of data sets is not guaranteed. So, e.g., if you
    have time series data, you should use caution.

    Example
    -------
    >>> import io
    >>> import pandas as pd
    >>> import bebi103
    >>> df = pd.read_csv(io.StringIO('''
        day,batch,colony,x
        monday,1,1,9.31
        monday,1,1,8.35
        monday,1,1,10.48
        monday,1,1,9.91
        monday,1,1,10.43
        monday,1,2,9.98
        monday,1,2,9.76
        monday,1,3,9.30
        monday,2,1,10.56
        monday,2,1,11.40
        monday,2,2,10.36
        monday,2,2,12.04
        monday,2,2,9.92
        monday,2,2,10.10
        monday,2,2,8.72
        monday,2,2,10.36
        monday,2,2,11.56
        monday,2,2,10.87
        monday,2,2,10.43
        monday,2,2,10.67
        monday,2,2,9.05
        monday,3,1,10.32
        monday,3,1,9.07
        monday,4,1,9.86
        monday,4,1,9.21
        monday,4,1,11.36
        monday,4,2,8.60
        monday,4,2,10.54
        monday,4,2,8.93
        monday,4,2,9.43
        monday,4,2,9.23
        monday,4,2,9.66
        monday,4,2,11.26
        monday,4,2,9.61
        monday,4,2,11.99
        monday,4,2,10.27
        monday,4,2,9.97
        monday,4,2,9.37
        monday,4,2,10.10
        monday,4,3,10.39
        monday,4,3,8.79
        wednesday,1,1,10.76
        wednesday,1,2,10.72
        wednesday,1,2,8.97
        wednesday,1,2,9.14
        wednesday,1,2,11.31
        wednesday,1,2,9.49
        wednesday,1,2,10.21
        wednesday,1,2,10.04
        wednesday,2,1,13.16
        wednesday,2,1,7.07
        wednesday,2,1,12.74
        wednesday,3,1,9.45
        wednesday,3,1,9.62
        wednesday,3,1,10.46
        wednesday,3,1,11.11
        wednesday,3,1,10.56
        wednesday,3,1,9.93
        thursday,1,1,8.60
        thursday,1,2,11.24
        thursday,1,2,9.10
        thursday,1,2,9.10
        thursday,1,2,11.30
        thursday,1,2,10.65
        thursday,1,2,9.98
        thursday,1,2,9.85
        thursday,1,2,12.41
        thursday,1,3,10.03
        thursday,1,3,10.53
        thursday,1,4,10.85
        '''), skipinitialspace=True)
    >>> data, df = bebi103.stan.df_to_datadict_hier(df,
                                level_cols=['day', 'batch', 'colony'],
                                data_cols=['x'])
    >>> data
    {'N': 70,
     'J_1': 3,
     'J_2': 8,
     'J_3': 17,
     'index_1': array([1, 1, 1, 1, 2, 3, 3, 3]),
     'index_2': array([1, 1, 1, 2, 2, 3, 4, 4, 4, 5, 5, 5, 5, 6, 6, 7, 8]),
     'index_3': array([ 1,  1,  1,  1,  1,  2,  2,  3,  4,  4,  5,  5,  5,  5,  5,  5,  5,
             5,  5,  5,  5,  6,  6,  7,  7,  7,  8,  8,  8,  8,  8,  8,  8,  8,
             8,  8,  8,  8,  8,  9,  9, 10, 11, 11, 11, 11, 11, 11, 11, 11, 12,
            12, 13, 14, 15, 15, 15, 15, 15, 15, 15, 16, 16, 16, 17, 17, 17, 17,
            17, 17]),
     'x': array([ 9.31,  8.35, 10.48,  9.91, 10.43,  9.98,  9.76,  9.3 , 10.56,
            11.4 , 10.36, 12.04,  9.92, 10.1 ,  8.72, 10.36, 11.56, 10.87,
            10.43, 10.67,  9.05, 10.32,  9.07,  9.86,  9.21, 11.36,  8.6 ,
            10.54,  8.93,  9.43,  9.23,  9.66, 11.26,  9.61, 11.99, 10.27,
             9.97,  9.37, 10.1 , 10.39,  8.79,  8.6 , 11.24,  9.1 ,  9.1 ,
            11.3 , 10.65,  9.98,  9.85, 12.41, 10.03, 10.53, 10.85, 10.76,
            10.72,  8.97,  9.14, 11.31,  9.49, 10.21, 10.04, 13.16,  7.07,
            12.74,  9.45,  9.62, 10.46, 11.11, 10.56,  9.93])}

    >>> df.head(10)
          day  batch  colony      x  day_stan  batch_stan  colony_stan
    0  monday      1       1   9.31         1           1            1
    1  monday      1       1   8.35         1           1            1
    2  monday      1       1  10.48         1           1            1
    3  monday      1       1   9.91         1           1            1
    4  monday      1       1  10.43         1           1            1
    5  monday      1       2   9.98         1           1            2
    6  monday      1       2   9.76         1           1            2
    7  monday      1       3   9.30         1           1            3
    8  monday      2       1  10.56         1           2            4
    9  monday      2       1  11.40         1           2            4
    """
    if df is None or level_cols is None or data_cols is None:
        raise RuntimeError("`df`, `level_cols`, and `data_cols` must all be specified.")

    if type(sort_cols) != list:
        raise RuntimeError("`sort_cols` must be a list.")

    # For now, if polars, just convert to pandas
    if 'polars.dataframe.frame.DataFrame' in str(type(df)):
        polars_in = True
        df = df.to_pandas()
    else:
        polars_in = False

    # Get a copy so we don't overwrite
    new_df = df.copy(deep=True)

    if type(level_cols) not in [list, tuple]:
        level_cols = [level_cols]

    if type(data_cols) not in [list, tuple]:
        data_cols = [data_cols]

    level_cols_stan = [col + "_stan" for col in level_cols]

    if cowardly:
        for col in level_cols_stan:
            if col in df:
                raise RuntimeError(
                    "column "
                    + col
                    + " already in data frame. Cowardly deciding not to overwrite."
                )

    for col_ind, col in enumerate(level_cols):
        new_df[str(col) + "_stan"] = df.groupby(level_cols[: col_ind + 1]).ngroup() + 1

    new_df = new_df.sort_values(by=level_cols_stan + sort_cols)

    data = dict()
    data["N"] = len(new_df)
    for i, col in enumerate(level_cols_stan):
        data["J_" + str(i + 1)] = len(new_df[col].unique())
    for i, _ in enumerate(level_cols_stan[1:]):
        data["index_" + str(i + 1)] = np.array(
            [key[i] for key in new_df.groupby(level_cols_stan[: i + 2]).groups]
        ).astype(int)
    data["index_" + str(len(level_cols_stan))] = new_df[
        level_cols_stan[-1]
    ].values.astype(int)
    for col in data_cols:
        # Check string naming
        new_col = str(col)
        try:
            bytes(new_col, "ascii")
        except UnicodeEncodeError:
            raise RuntimeError("Column names must be ASCII.")
        if new_col[0].isdigit():
            raise RuntimeError("Column name cannot start with a number.")
        for char in new_col:
            if char in "`~!@#$%^&*()- =+[]{}\\|:;\"',<.>/?":
                raise RuntimeError("Invalid column name for Stan variable.")

        data[new_col] = new_df[col].values

    # Return polars if input was polars
    if polars_in:
        new_df = pl.from_pandas(new_df)

    return data, new_df


def arviz_to_dataframe(
    data, var_names=None, diagnostics=("diverging",), df_package="polars"
):
    """Convert ArviZ InferenceData to a Pandas or Polars data frame.

    Any multi-dimensional parameters are converted to one-dimensional
    equivalents. For example, a 2x2 matrix A is converted to columns
    in the data frame with headings `'A[0,0]'`, `'A[0,1]'`, `'A[1,0]'`,
    and `'A[1,1]'`.

    Only divergence information, chain ID, and draw number (not
    additional sampling stats) are stored in the data frame.

    Parameters
    ----------
    data : ArviZ InferenceData instance
        Results from MCMC sampling to convert to a data frame. It must
        have a `posterior` attribute, and it should have a
        `sample_stats` attribute if divergence information is to be
        stored.
    var_names : list of strings or None
        List of variables present in `data.posterior.data_vars` to store
        in the data frame.  If None, all variables present in
        `data.posterior.data_vars` are stored. Note that *only*
        variables in `data.posterior.data_vars` may be stored.
    diagnostics : list or tuple of strings
        Diagnostic data to be stored in the data frame. The elements of
        the list may include: 'lp', 'acceptance_rate', 'step_size',
        'tree_depth', 'n_steps', 'diverging', and 'energy'.
    df_package : str, one of 'polars' (default) or 'pandas'
        What type of data frame to output

    Returns
    -------
    output : Polars or Pandas DataFrame
        DataFrame with posterior samples and diagnostics. The column
        names of all diagnostics, as well as 'chain' and 'draw', are
        appended with a double-underscore (`__`) to signify that they
        are not samples out of the posterior, but rather metadata about
        a sample.

    """
    if type(data) == xarray.Dataset:
        raise RuntimeError(
            "You need to pass in an ArviZ InferenceData instance, not an xarray Dataset. Maybe you passed in the posterior attribute of the InferenceData instance?"
        )

    if diagnostics is not None:
        if type(diagnostics) == str:
            diagnostics = (diagnostics,)

    var_names = _parameters_to_arviz_var_names(data, var_names)

    if not hasattr(data, "posterior"):
        raise RuntimeError("No posterior contained in `data`.")

    diag_dict = {}
    if diagnostics is not None and len(diagnostics) > 0:
        if not hasattr(data, "sample_stats"):
            raise RuntimeError(
                "Asking for diagnostics, but input has not attribute sample_stats"
            )
        for diag in diagnostics:
            if hasattr(data.sample_stats, diag):
                diag_dict[diag + "__"] = np.ravel(data.sample_stats[diag])
            else:
                raise RuntimeError(f"{diag} not in data.sample_stats.")

    cols, data_as_ndarray = _xarray_to_ndarray(data.posterior, var_names=var_names)

    chain = np.concatenate(
        [
            [i] * data.posterior.sizes["draw"]
            for i in range(data.posterior.sizes["chain"])
        ]
    )

    draw = np.concatenate(
        [data.posterior["draw"].values for i in range(data.posterior.sizes["chain"])]
    )

    if type(df_package) == str and df_package.lower() == "pandas":
        df = pd.DataFrame(data=data_as_ndarray.T, columns=cols)
        df["chain__"] = chain
        df["draw__"] = draw
        for diag, res in diag_dict.items():
            df[diag] = res
    elif type(df_package) == str and df_package.lower() == "polars":
        df = pl.DataFrame(data=data_as_ndarray.T, schema=cols)
        df = df.with_columns(
            pl.Series(chain).alias("chain__"),
            pl.Series(draw).alias("draw__"),
            *[pl.Series(res).alias(diag) for diag, res in diag_dict.items()],
        )
    else:
        raise RuntimeError("Invalid `df_package`. Must be either 'polars' or 'pandas'.")

    return df


def check_divergences(samples, quiet=False, return_diagnostics=False):
    """Check transitions that ended with a divergence.

    Parameters
    ----------
    samples : ArviZ InferenceData instance
        Contains samples to be checked. Must contain both `posterior`
        and `sample_stats`.
    quiet : bool, default False
        If True, do not print diagnostic result to the screen.
    return_diagnostics : bool, default False
        If True, return both a Boolean about whether the diagnostic
        passed and the number of samples where the tree depth was too
        deep. Otherwise, only return Boolean if the test passed.

    Returns
    -------
    passed : bool
        Return True if tree depth test passed. Return False otherwise.
    n_divergent : int, optional
        Number of divergent samples.
    """
    n_divergent = samples.sample_stats["diverging"].values.sum()
    n_total = samples.sample_stats.sizes["chain"] * samples.sample_stats.sizes["draw"]

    if not quiet:
        msg = "{} of {} ({}%) iterations ended with a divergence.".format(
            n_divergent, n_total, 100 * n_divergent / n_total
        )
        print(msg)

    pass_check = n_divergent == 0

    if not pass_check and not quiet:
        print("  Try running with larger adapt_delta to remove divergences.")

    if return_diagnostics:
        return pass_check, n_divergent
    return pass_check


def check_treedepth(samples, max_treedepth=10, quiet=False, return_diagnostics=False):
    """Check transitions that ended prematurely due to maximum tree depth limit.

    Parameters
    ----------
    samples : ArviZ InferenceData instance
        Contains samples to be checked. Must contain both `posterior`
        and `sample_stats`.
    max_treedepth: int, default 10
        Maximum tree depth used in the calculation.
    quiet : bool, default False
        If True, do not print diagnostic result to the screen.
    return_diagnostics : bool, default False
        If True, return both a Boolean about whether the diagnostic
        passed and the number of samples where the tree depth was too
        deep. Otherwise, only return Boolean if the test passed.

    Returns
    -------
    passed : bool
        Return True if tree depth test passed. Return False otherwise.
    n_too_deep : int, optional
        Number of samplers wherein the tree depth was greater than
        `max_treedepth`.
    """
    # ArviZ v. 0.11.2 updated treedepth to be tree_depth and v. 0.11.4 reverted
    try:
        n_too_deep = (samples.sample_stats.tree_depth.values >= max_treedepth).sum()
    except:
        n_too_deep = (samples.sample_stats.treedepth.values >= max_treedepth).sum()

    n_total = samples.sample_stats.sizes["chain"] * samples.sample_stats.sizes["draw"]

    if not quiet:
        msg = "{} of {} ({}%) iterations saturated".format(
            n_too_deep, n_total, 100 * n_too_deep / n_total
        )
        msg += " the maximum tree depth of {}.".format(max_treedepth)
        print(msg)

    pass_check = n_too_deep == 0

    if not pass_check and not quiet:
        print(
            "  Try running again with max_treedepth set to a larger value"
            + " to avoid saturation."
        )

    if return_diagnostics:
        return pass_check, n_too_deep
    return pass_check


def check_energy(
    samples, e_bfmi_rule_of_thumb=0.3, quiet=False, return_diagnostics=False
):
    """Checks the energy-Bayes fraction of missing information (E-BFMI)

    Parameters
    ----------
    samples : ArviZ InferenceData instance
        Contains samples to be checked. Must contain both `posterior`
        and `sample_stats`.
    e_bfmi_rule_of_thumb : float, default 0.3 (as per cmdstan)
        Rule of thumb value for E-BFMI. If below this value, there may
        be cause for concern.
    quiet : bool, default False
        If True, do no print diagnostic result to the screen.
    return_diagnostics : bool, default False
        If True, return both a Boolean about whether the diagnostic
        passed and a data frame containing results about the E-BMFI
        tests. Otherwise, only return Boolean if the test passed.

    Returns
    -------
    passed : bool
        Return True if test passed. Return False otherwise.
    e_bfmi_diagnostics : DataFrame
        DataFrame with information about which chains had problematic
        E-BFMIs.
    """
    # Alternatively:
    # samples.sample_stats["energy"].groupby("chain").apply(_ebfmi).values
    ebfmi = az.bfmi(samples)

    problematic = ebfmi < e_bfmi_rule_of_thumb

    pass_check = (~problematic).all()

    if not quiet:
        if pass_check:
            print("E-BFMI indicated no pathological behavior.")
        else:
            for i, (problem, ebfmi_val) in enumerate(zip(problematic, ebfmi)):
                print("Chain {}: E-BFMI = {}".format(i, ebfmi_val))
            print(
                "  E-BFMI below 0.3 indicates you may need to "
                + "reparametrize your model."
            )

    if return_diagnostics:
        return (
            pass_check,
            pd.DataFrame(
                {
                    "chain": np.arange(len(ebfmi)),
                    "E-BFMI": ebfmi,
                    "problematic": ebfmi < e_bfmi_rule_of_thumb,
                }
            ),
        )
    return pass_check


def check_ess(
    samples,
    parameters=None,
    total_ess_rule_of_thumb=100,
    fractional_ess_rule_of_thumb=0.001,
    quiet=False,
    return_diagnostics=False,
):
    """Checks the effective sample size (ESS).

    Parameters
    ----------
    samples : ArviZ InferenceData instance
        Contains samples to be checked. Must contain both `posterior`
        and `sample_stats`.
    parameters : list of str, or None (default)
        Names of parameters to use. If None, use all parameters. For
        multidimensional parameters, each entry must be given
        separately, e.g., `['alpha[0]', 'alpha[1]', 'beta[0,1]']`.
    quiet : bool, default False
        If True, do not print diagnostic result to the screen.
    total_ess_rule_of_thumb : int, default 100
        Rule of thumb for number of effective samples per chain. The
        default of 100 is based on the suggestion of Vehtari, et al., to
        have 50 effective samples per split chain.
    fractional_ess_rule_of_thumb : float, default 0.001
        Rule of thumb value for fractional number of effective samples.
        The default of 0.001 is based on CmdStan's defaults.
    return_diagnostics : bool, default False
        If True, return both a Boolean about whether the diagnostic
        passed and a data frame containing results about the number
        of effective samples tests. Otherwise, only return Boolean if
        the test passed.

    Returns
    -------
    passed : bool
        Return True if test passed. Return False otherwise.
    ess_diagnostics : DataFrame
        Data frame with information about problematic ESS.

    Notes
    -----
    Parameters with ESS given as NaN are not checked.

    References
    ----------
    Vehtari, et al., 2019, https://arxiv.org/abs/1903.08008.
    """
    # For convenience
    N = samples.posterior.sizes["draw"]
    M = samples.posterior.sizes["chain"]
    total_ess_rule_of_thumb *= M

    var_names_arviz = _parameters_to_arviz_var_names(samples, parameters)

    ess = az.ess(samples, var_names=var_names_arviz, method="bulk")
    tail_ess = az.ess(samples, var_names=var_names_arviz, method="tail")

    # Convert to list of names and numpy arrays
    names, ess = _xarray_to_ndarray(ess)
    _, tail_ess = _xarray_to_ndarray(tail_ess)
    if ess.shape == (1, 1):
        ess = np.array([ess[0, 0]])
        tail_ess = np.array([tail_ess[0, 0]])
    else:
        ess = ess.squeeze()
        tail_ess = tail_ess.squeeze()

    # Fractional ESS
    frac_ess = ess / M / N
    frac_tail_ess = tail_ess / M / N

    pass_check = (
        (ess[~np.isnan(ess)] > total_ess_rule_of_thumb).all()
        and (frac_ess[~np.isnan(ess)] > fractional_ess_rule_of_thumb).all()
        and (tail_ess[~np.isnan(tail_ess)] > total_ess_rule_of_thumb).all()
        and (frac_tail_ess[~np.isnan(tail_ess)] > fractional_ess_rule_of_thumb).all()
    )

    if not quiet:
        if not pass_check:
            n_e = 0
            n_f = 0
            for name, e, f, te, tf in zip(
                names, ess, frac_ess, tail_ess, frac_tail_ess
            ):
                if e < total_ess_rule_of_thumb:
                    print("ESS for parameter {} is {}.".format(name, e))
                    n_e += 1
                if f < fractional_ess_rule_of_thumb:
                    print("ESS / iter for parameter {} is {}.".format(name, f))
                    n_f += 1
                if te < total_ess_rule_of_thumb:
                    print("tail-ESS for parameter {} is {}.".format(name, te))
                    n_e += 1
                if tf < fractional_ess_rule_of_thumb:
                    print("ESS / iter for parameter {} is {}.".format(name, tf))
                    n_f += 1
            if n_e > 0:
                print(
                    """  ESS or tail-ESS below 100 per chain indicates that expectation values
  computed from samples are unlikely to be good approximations of the
  true expectation values."""
                )
            if n_f > 0:
                print(
                    """  ESS / iter or tail-ESS / iter below 0.001 indicates that the effective
  sample size has likely been overestimated."""
                )
        else:
            print("Effective sample size looks reasonable for all parameters.")

    if return_diagnostics:
        return (
            pass_check,
            pd.DataFrame(
                data={
                    "parameter": names,
                    "ESS": ess,
                    "ESS_per_iter": frac_ess,
                    "tail_ESS": tail_ess,
                    "tail_ESS_per_iter": frac_tail_ess,
                }
            ),
        )
    return pass_check


def check_rhat(
    samples,
    parameters=None,
    rhat_rule_of_thumb=1.01,
    omit=(),
    quiet=False,
    return_diagnostics=False,
):
    """Checks the potential issues with scale reduction factors. Rhat
    is computed with rank-normalization and folding.

    Parameters
    ----------
    samples : ArviZ InferenceData instance
        Contains samples to be checked. Must contain both `posterior`
        and `sample_stats`.
    parameters : list of str, or None (default)
        Names of parameters to use. If None, use all parameters. For
        multidimensional parameters, each entry must be given
        separately, e.g., `['alpha[0]', 'alpha[1]', 'beta[0,1]']`.
    quiet : bool, default False
        If True, do no print diagnostic result to the screen.
    rhat_rule_of_thumb : float, default 1.01
        Rule of thumb value for maximum allowed R-hat, as per Vehtari,
        et al.
    omit : str, re.Pattern, or list or tuple of str and re.Pattern
        If `parameters` is not provided, all sampled parameters are
        checked for Rhat. We often want to ignore samples of variables
        that a transformed parameters, as their Rhats are irrelevant and
        often NaNs. For each string entry in `omit`, the variable given
        by the string is omitted. For each entry that is a compiled
        regular expression patters (`re.Pattern`), any variable name
        matching the pattern is omitted. By default, no variables are
        omitted.
    return_diagnostics : bool, default False
        If True, return both a Boolean about whether the diagnostic
        passed and a data frame containing results about the number
        of effective samples tests. Otherwise, only return Boolean if
        the test passed.

    Returns
    -------
    passed : bool
        Return True if test passed. Return False otherwise.
    rhat_diagnostics : DataFrame
        Data frame with information about problematic R-hat values.

    References
    ----------
    Vehtari, et al., 2019, https://arxiv.org/abs/1903.08008.

    """
    if omit is None:
        omit = []
    elif type(omit) not in [tuple, list]:
        omit = [omit]

    var_names_arviz = _parameters_to_arviz_var_names(samples, parameters)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        rhat = az.rhat(samples, var_names=var_names_arviz, method="rank")

    # Convert to list of names and numpy arrays
    names, rhat = _xarray_to_ndarray(rhat)

    if rhat.shape == (1, 1):
        rhat = np.array([rhat[0, 0]])
    else:
        rhat = rhat.squeeze()

    if parameters is None:
        rhat = np.array(
            [r for name, r in zip(names, rhat) if _screen_param(name, omit)]
        )
        names = [name for name in names if _screen_param(name, omit)]
    else:
        rhat = np.array([r for name, r in zip(names, rhat) if name in parameters])
        names = [name for name, r in zip(names, rhat) if name in parameters]

    nans = np.isnan(rhat)
    non_nan_pass = np.sum(np.array(rhat[~nans]) > rhat_rule_of_thumb) == 0
    nan_pass = nans.sum() == 0

    pass_check = non_nan_pass and nan_pass

    if not quiet:
        if not pass_check:
            if not nan_pass:
                for name, nan in zip(names, nans):
                    if nan:
                        print("Rhat for parameter {} is NaN.".format(name))

            if not non_nan_pass:
                for name, r in zip(names, rhat):
                    if r > rhat_rule_of_thumb:
                        print("Rhat for parameter {} is {}.".format(name, r))
                print(
                    "  Rank-normalized Rhat above 1.01 indicates that the chains very"
                    " likely have not mixed."
                )
        else:
            print("Rhat looks reasonable for all parameters.")

    if return_diagnostics:
        return pass_check, pd.DataFrame(data={"parameter": names, "Rhat": rhat})

    return pass_check


def check_all_diagnostics(
    samples,
    parameters=None,
    e_bfmi_rule_of_thumb=0.3,
    total_ess_rule_of_thumb=100,
    fractional_ess_rule_of_thumb=0.001,
    rhat_rule_of_thumb=1.01,
    omit=(),
    max_treedepth=10,
    quiet=False,
    return_diagnostics=False,
):
    """Checks all MCMC diagnostics

    Parameters
    ----------
    samples : ArviZ InferenceData instance
        Contains samples to be checked. Must contain both `posterior`
        and `sample_stats`.
    parameters : list of str, or None (default)
        Names of parameters to use in checking Rhat and ESS. If None,
        use all parameters. For multidimensional parameters, each entry
        must be given separately, e.g.,
        `['alpha[0]', 'alpha[1]', 'beta[0,1]']`.
    e_bfmi_rule_of_thumb : float, default 0.3
        Rule of thumb value for E-BFMI. If below this value, there may
        be cause for concern.
    total_ess_rule_of_thumb : int, default 100
        Rule of thumb for number of effective samples per chain. The
        default of 100 is based on the suggestion of Vehtari, et al., to
        have 50 effective samples per split chain.
    fractional_ess_rule_of_thumb : float, default 0.001
        Rule of thumb value for fractional number of effective samples.
        The default of 0.001 is based on CmdStan's defaults.
    rhat_rule_of_thumb : float, default 1.1
        Rule of thumb value for maximum allowed R-hat.
    omit : str, re.Pattern, or list or tuple of str and re.Pattern
        If `parameters` is not provided, all sampled parameters are
        checked for Rhat. We often want to ignore samples of variables
        that a transformed parameters, as their Rhats are irrelevant and
        often NaNs. For each string entry in `omit`, the variable given
        by the string is omitted. For each entry that is a compiled
        regular expression patters (`re.Pattern`), any variable name
        matching the pattern is omitted. By default, no variables are
        omitted.
    max_treedepth: int, default 'infer'
        If int, specification of maximum treedepth allowed. If 'infer',
        inferred from `fit`.
    quiet : bool, default False
        If True, do no print diagnostic result to the screen.
    return_diagnostics : bool, default False
        If True, return a dictionary containing the results of each test.

    Returns
    -------
    warning_code : int
        When converted to binary, each digit in the code stands for
        whether or not a test passed. A digit of zero indicates the test
        passed. The ordering of the tests goes:

        - ess
        - r_hat
        - divergences
        - tree depth
        - E-BFMI

        For example, a warning code of 12 has a binary representation
        of 01100, which means that R-hat and divergences tests failed.
    return_dict : dict
        Returned if `return_dict` is True. A dictionary with the result
        of each diagnostic test.
    """
    warning_code = 0
    diag_results = {}

    pass_check, diag_results["ess"] = check_ess(
        samples,
        parameters=parameters,
        total_ess_rule_of_thumb=total_ess_rule_of_thumb,
        fractional_ess_rule_of_thumb=fractional_ess_rule_of_thumb,
        quiet=quiet,
        return_diagnostics=True,
    )
    if not pass_check:
        warning_code = warning_code | (1 << 0)

    if not quiet:
        print("")

    pass_check, diag_results["rhat"] = check_rhat(
        samples,
        parameters=parameters,
        rhat_rule_of_thumb=rhat_rule_of_thumb,
        omit=omit,
        quiet=quiet,
        return_diagnostics=True,
    )
    if not pass_check:
        warning_code = warning_code | (1 << 1)

    if not quiet:
        print("")

    pass_check, diag_results["n_divergences"] = check_divergences(
        samples, quiet=quiet, return_diagnostics=True
    )
    if not pass_check:
        warning_code = warning_code | (1 << 2)

    if not quiet:
        print("")

    pass_check, diag_results["n_max_treedepth"] = check_treedepth(
        samples, max_treedepth=max_treedepth, quiet=quiet, return_diagnostics=True
    )
    if not pass_check:
        warning_code = warning_code | (1 << 3)

    if not quiet:
        print("")

    pass_check, diag_results["e_bfmi"] = check_energy(
        samples,
        e_bfmi_rule_of_thumb=e_bfmi_rule_of_thumb,
        quiet=quiet,
        return_diagnostics=True,
    )
    if not pass_check:
        warning_code = warning_code | (1 << 4)

    if return_diagnostics:
        return warning_code, diag_results

    return warning_code


def parse_warning_code(warning_code, quiet=False, return_dict=False):
    """Parses warning code from `check_all_diagnostics()` into
    individual failures and prints results.

    Parameters
    ----------
    warning_code : int
        When converted to binary, each digit in the code stands for
        whether or not a test passed. A digit of zero indicates the test
        passed. The ordering of the tests goes:

        - ESS
        - r_hat
        - divergences
        - tree depth
        - E-BFMI

        For example, a warning code of 12 has a binary representation
        of 01100, which means that R-hat and divergences tests failed.
    quiet : bool, default False
        If True, suppress print results to the screen.
    return_dict : bool, default False
        If True, return a dictionary of containing test passage info.

    Returns
    -------
    output : dict
        If `return_dict` is True, returns a dictionary where each entry
        is True is the respective diagnostic check was passed and False
        if it was not.
    """

    if quiet and not return_dict:
        raise RuntimeError(
            "`quiet` is True and `return_dict` is False, "
            + "so there is nothing to do."
        )

    passed_tests = dict(
        neff=True, rhat=True, divergence=True, treedepth=True, energy=True
    )

    if warning_code & (1 << 0):
        passed_tests["neff"] = False
        if not quiet:
            print("ESS warning")
    if warning_code & (1 << 1):
        passed_tests["rhat"] = False
        if not quiet:
            print("Rhat warning")
    if warning_code & (1 << 2):
        passed_tests["divergence"] = False
        if not quiet:
            print("divergence warning")
    if warning_code & (1 << 3):
        passed_tests["treedepth"] = False
        if not quiet:
            print("treedepth warning")
    if warning_code & (1 << 4):
        passed_tests["energy"] = False
        if not quiet:
            print("energy warning")
    if warning_code == 0:
        if not quiet:
            print("No diagnostic warnings")

    if return_dict:
        return passed_tests


def _multiprocessing_spawn_context():
    """
    Return a multiprocessing spawn context to avoid fork deadlocks
    """
    try:
        return multiprocessing.get_context("spawn")
    except Exception:
        return multiprocessing.get_context()


def sbc(
    prior_predictive_stan_file=None,
    posterior_stan_file=None,
    prior_predictive_model_data=None,
    posterior_model_data=None,
    measured_data=None,
    var_names=None,
    measured_data_dtypes=None,
    posterior_predictive_var_names=None,
    log_likelihood_var_name=None,
    sampling_kwargs=None,
    diagnostic_check_kwargs=None,
    cores=1,
    N=400,
    n_prior_draws_for_sd=1000,
    samples_dir="sbc_samples",
    remove_sample_files=True,
    df_package='polars',
    progress_bar=False,
    output_csv='sbc_results.csv',
    overwrite=False,
):
    """Perform simulation-based calibration on a Stan Model.

    Parameters
    ----------
    prior_predictive_stan_file : str
        Stan file for generating prior predictive data sets.
    posterior_model : str
        Stan file for the posterior that allows sampling.
    prior_predictive_model_data : dict
        Dictionary with entries specified by the data block of the prior
        predictive Stan model.
    posterior_model_data : dict
        Dictionary with entries specified by the data block of the prior
        predictive Stan model. Measured data in this dictionary will be
        replaced in each simulation by what was generated by the prior
        predictive model.
    measured_data : list
        A list of strings containing the variable names of measured
        data. Each entry in `measured_data` must be a key in
        `posterior_model_data`.
    var_names : list of strings
        A list of strings containing parameter names to be considered
        in the SBC analysis. Not all parameters of the model need be
        considered; only those in `var_names` have rank statistics
        calculated. Note that for multidimensional
        variables, `var_names` only has the root name. E.g.,
        `var_names=['x', 'y']`, *not* something like
        `var_names=['x[0]', 'x[1]', 'y[0,0]']`.
    posterior_predictive_var_names : list of strings, default None
        List of variables that are posterior predictive. These are
        ignored in the SBC analysis. Note that for multidimensional
        variables, `var_names` only has the root name. E.g.,
        `var_names=['x_ppc', 'y_ppc']`, *not* something like
        `var_names=['x_ppc[0]', 'x_ppc[1]', 'y_ppc[0,0]']`.
    log_likelihood_var_name : string, default None
        Name of variable in the Stan model that stores the log
        likelihood. This is ignored in the SBC analysis.
    measured_data_dtypes : dict, default None
        The key in the dtypes dict is a string representing the data
        name, and the corresponding item is its dtype, almost always
        either `int` or `float`.
    sampling_kwargs : dict, default None
        kwargs to be passed to `sm.sample()`.
    diagnostic_check_kwargs : dict, default None
        kwargs to pass to `check_all_diagnostics()`. If `quiet` and/or
        `return_diagnostics` are given, they are ignored.
        `max_treedepth` is inferred from `sampling_kwargs`.
    cores : int, default 1
        Number of cores to use in the SBC calculation.
    N : int, 400
        Number of simulations to run.
    n_prior_draws_for_sd : int, default 1000
        Number of prior draws to compute the prior standard deviation
        for a parameter in the prior distribution. This standard
        deviation is used in the shrinkage calculation.
    samples_dir : str, default "sbc_samples"
        Path to directory to store .csv and .txt files generated by
        CmdStan. The directory specified here will NOT be destroyed 
        after a calculation, though the files within it may be, 
        depending on the `remove_sample_files` kwarg.
    remove_sample_files : bool, default True
        If True, remove .csv and .txt files generated by CmdStan.
    df_package : str, either 'polars' (default) or 'pandas'
        Which package to use for output data frame
    progress_bar : bool, default False
        If True, display a progress bar for the calculation using tqdm.
    output_csv : str, default 'sbc_results.csv'
        Path to CSV file where results will be written. Results are
        written directly to this file as workers complete (instead of
        collecting in memory). This reduces memory usage and allows
        monitoring progress by examining the growing CSV file. The file
        includes metadata comments at the top with run parameters and
        start time.
    overwrite : bool, default False
        If False, raises RuntimeError if output_csv already exists.
        If True, overwrites existing file with new results.

    Returns
    -------
    output : Polars or Pandas DataFrame
        Data frame with the output of the SBC analysis. It has
        the following columns.

            - trial : Unique trial number for the simulation.
            - warning_code : Warning code based on diagnostic checks outputted by `check_all_diagnostics()`.
            - parameter: The name of the scalar parameter.
            - prior: Value of the parameter used in the simulation. This value was drawn out of the prior distribution.
            - mean : mean parameter value based on sampling out of the posterior in the simulation.
            - sd : standard deviation of the parameter value based on sampling out of the posterior in the simulation.
            - L : The number of bins used in computing the rank statistic. The rank statistic should be uniform on the integers [0, L].
            - rank_statistic : Value of the rank statistic for the parameter for the trial.
            - shrinkage : The shrinkage for the parameter for the given trial. This is computed as 1 - sd / sd_prior, where sd_prior is the standard deviation of the parameters as determined from drawing out of the prior.
            - z_score : The z-score for the parameter for the given trial. This is computed as abs(mean - prior) / sd.

    Notes
    -----
    Each simulation is done by sampling a parameter set out of the
    prior distribution, using those parameters to generate data from
    the likelihood, and then performing posterior sampling based on
    the generated data. A rank statistic for each simulation is
    computed. This rank statistic should be uniformly distributed
    over its L possible values. See https://arxiv.org/abs/1804.06788,
    by Talts, et al., for details.

    """
    if measured_data_dtypes is None:
        measured_data_dtypes = {}

    if sampling_kwargs is None:
        sampling_kwargs = {}

    if diagnostic_check_kwargs is None:
        diagnostic_check_kwargs = {}

    if prior_predictive_stan_file is None:
        raise RuntimeError("`prior_predictive_stan_file` must be specified.")
    if posterior_stan_file is None:
        raise RuntimeError("`posterior_stan_file` must be specified.")
    if prior_predictive_model_data is None:
        raise RuntimeError("`prior_predictive_model_data` must be specified.")
    if posterior_model_data is None:
        raise RuntimeError("`posterior_model_data` must be specified.")
    if measured_data is None:
        raise RuntimeError("`measured_data` must be specified.")

    if "output_dir" in sampling_kwargs:
        raise RuntimeError(
            "The 'output_dir' kwarg is not allowed because unambiguous naming is not possible if new sampling is done more than once per minute, given CmdStanPy's naming convention."
        )

    # Defaults for diagnostic checks
    diagnostic_check_kwargs["quiet"] = True
    diagnostic_check_kwargs["return_diagnostics"] = True
    if (
        "max_treedepth" not in diagnostic_check_kwargs
        and "max_treedepth" in sampling_kwargs
    ):
        diagnostic_check_kwargs["max_treedepth"] = sampling_kwargs["max_treedepth"]

    # We parallelize simulations, not chains within each simulation
    if "n_jobs" in sampling_kwargs:
        del sampling_kwargs["n_jobs"]
    if "parallel_chains" in sampling_kwargs:
        del sampling_kwargs["parallel_chains"]

    # Shut off CmdStanPy's progress bar by default
    if "show_progress" not in sampling_kwargs:
        sampling_kwargs["show_progress"] = False    
    
    # Compile prior predictive model so we can set up data types, etc.
    prior_predictive_model = cmdstanpy.CmdStanModel(stan_file=prior_predictive_stan_file)

    # Take a prior sample to infer data types
    with disable_logging():
        prior_sample = prior_predictive_model.sample(
            data=prior_predictive_model_data,
            fixed_param=True,
            chains=1,
            iter_sampling=1,
            show_progress=False,
        )
    prior_sample = az.from_cmdstanpy(
        prior=prior_sample, prior_predictive=measured_data
    )

    # Infer dtypes of measured data
    for data in measured_data:
        ar = prior_sample.prior_predictive[data].squeeze()
        if data not in measured_data_dtypes:
            if np.sum(ar != ar.astype(int)) == 0:
                warnings.warn(f"Inferring int dtype for {data}.")
                measured_data_dtypes[data] = int
            else:
                measured_data_dtypes[data] = float

    # Check parameters
    if var_names is None:
        var_names = [
            param
            for param in prior_sample.prior.data_vars
            if len(param) < 2 or param[-2:] != "__"
        ]
    else:
        for param in var_names:
            if param not in prior_sample.prior.data_vars:
                raise RuntimeError(
                    f"parameter `{param}` not in prior sample generated from `prior_predictive_model`."
                )

    # Determine prior SDs for parameters of interest
    prior_sd = _get_prior_sds(
        prior_predictive_model,
        prior_predictive_model_data,
        var_names,
        measured_data,
        n_prior_draws_for_sd,
        samples_dir,
        remove_sample_files,
    )

    # Determine number of iterations
    thin = sampling_kwargs["thin"] if "thin" in sampling_kwargs else 1
    chains = sampling_kwargs["chains"] if "chains" in sampling_kwargs else 4
    if "iter_sampling" in sampling_kwargs:
        iter_sampling = sampling_kwargs["iter_sampling"]
    else:
        iter_sampling = 1000
    L = iter_sampling * chains // thin

    # Check if output file exists and handle overwrite
    if os.path.exists(output_csv) and not overwrite:
        raise RuntimeError(
            f"Output file '{output_csv}' already exists. "
            "Set overwrite=True to overwrite the existing file."
        )

    # Create manager for shared lock
    manager = multiprocessing.Manager()
    write_lock = manager.Lock()

    # Initialize CSV file with metadata comments
    start_time = datetime.datetime.now()
    with open(output_csv, 'w') as f:
        f.write(f"# SBC Analysis Results\n")
        f.write(f"# Start time: {start_time.strftime('%Y-%m-%d %H:%M:%S')}\n")
        f.write(f"#\n")
        f.write(f"# Parameters:\n")
        f.write(f"#   N: {N}\n")
        f.write(f"#   cores: {cores}\n")
        f.write(f"#   prior_predictive_stan_file: {prior_predictive_stan_file}\n")
        f.write(f"#   posterior_stan_file: {posterior_stan_file}\n")
        f.write(f"#   var_names: {var_names}\n")
        f.write(f"#   measured_data: {measured_data}\n")
        f.write(f"#   sampling_kwargs: {sampling_kwargs}\n")
        f.write(f"#   diagnostic_check_kwargs: {diagnostic_check_kwargs}\n")
        f.write(f"#   samples_dir: {samples_dir}\n")
        f.write(f"#   remove_sample_files: {remove_sample_files}\n")
        f.write(f"#\n")
        # Write CSV header
        f.write("trial,parameter,ground_truth,rank_statistic,mean,sd,shrinkage,z_score,Rhat,ESS,ESS_per_iter,tail_ESS,tail_ESS_per_iter,n_divergences,n_bad_ebfmi,n_max_treedepth,warning_code,L,error\n")

    def arg_input_generator():
        counter = 0
        while counter < N:
            yield (
                prior_predictive_stan_file,
                posterior_stan_file,
                prior_predictive_model_data,
                posterior_model_data,
                measured_data,
                var_names,
                measured_data_dtypes,
                sampling_kwargs,
                diagnostic_check_kwargs,
                posterior_predictive_var_names,
                log_likelihood_var_name,
                prior_sd,
                samples_dir,
                remove_sample_files,
                output_csv,
                write_lock,
                L,
                counter,  # trial number
            )
            counter += 1

    # Use multiprocessing with spawn context
    mp_context = _multiprocessing_spawn_context()
    with mp_context.Pool(cores) as pool:
        iterator = pool.imap_unordered(_perform_sbc, arg_input_generator(), chunksize=1)

        # Consume the iterator (results written to CSV by workers)
        if progress_bar == "notebook":
            for _ in tqdm.tqdm_notebook(iterator, total=N):
                pass
        elif progress_bar == True:
            for _ in tqdm.tqdm(iterator, total=N):
                pass
        elif progress_bar == False:
            for _ in iterator:
                pass
        else:
            raise RuntimeError("Invalid `progress_bar`.")

    # Read back the CSV and return as DataFrame
    if type(df_package) == str and df_package.lower() == 'polars':
        return pl.read_csv(output_csv, comment_prefix='#')
    else:
        return pd.read_csv(output_csv, comment='#')


def _perform_sbc(args):
    """Perform an SBC analysis"""
    (
        prior_predictive_stan_file,
        posterior_stan_file,
        prior_predictive_model_data,
        posterior_model_data,
        measured_data,
        var_names,
        measured_data_dtypes,
        sampling_kwargs,
        diagnostic_check_kwargs,
        posterior_predictive_var_names,
        log_likelihood_var_name,
        prior_sd,
        samples_dir,
        remove_sample_files,
        output_csv,
        write_lock,
        L,
        trial,
    ) = args

    posterior_model_data = copy.deepcopy(posterior_model_data)

    with disable_logging():
        output_dir = _get_output_dir(samples_dir, "prior_pred")

        # Build models locally within worker (should default to pre-build exe file)
        prior_predictive_model = cmdstanpy.CmdStanModel(stan_file=prior_predictive_stan_file)
        posterior_model = cmdstanpy.CmdStanModel(stan_file=posterior_stan_file)

        prior_sample_cmdstanpy = prior_predictive_model.sample(
            data=prior_predictive_model_data,
            fixed_param=True,
            chains=1,
            iter_sampling=1,
            show_progress=False,
            output_dir=output_dir,
        )
    prior_sample = az.from_cmdstanpy(
        prior=prior_sample_cmdstanpy, prior_predictive=measured_data
    )
    if remove_sample_files:
        _remove_sampling_files(prior_sample_cmdstanpy.runset)
    
    # Extract data generated from the prior predictive calculation
    for data in measured_data:
        ar = prior_sample.prior_predictive[data].squeeze().values
        if len(ar.shape) == 0:
            posterior_model_data[data] = ar.astype(measured_data_dtypes[data]).item()
        else:
            posterior_model_data[data] = ar.astype(measured_data_dtypes[data])

    # Store what the parameters were to generate prior predictive data
    names, vals = _xarray_to_ndarray(prior_sample.prior)
    param_priors = {name: val.item() for name, val in zip(names, vals)}

    # Generate posterior samples
    try:
        with disable_logging():
            output_dir = _get_output_dir(samples_dir, "model")
            posterior_samples_cmdstanpy = posterior_model.sample(
                data=posterior_model_data,
                parallel_chains=1,
                output_dir=output_dir,
                **sampling_kwargs,
            )
        posterior_samples = az.from_cmdstanpy(
            posterior=posterior_samples_cmdstanpy,
            posterior_predictive=posterior_predictive_var_names,
            log_likelihood=log_likelihood_var_name,
        )

        # Clean out samples to save disk space
        if remove_sample_files:
            _remove_sampling_files(posterior_samples_cmdstanpy.runset)

        # Check diagnostics
        warning_code, diagnostics = check_all_diagnostics(
            posterior_samples, **diagnostic_check_kwargs
        )
        err_str = 'no error'
        success = True
    except Exception as exception:
        err_str = exception.__str__()
        warnings.warn(f"Trial failure with error message: {err_str}")
        success = False

    if success:
        # Convert output to Numpy array
        names, vals = _xarray_to_ndarray(posterior_samples.posterior, var_names=var_names)

        # Generate output dictionary
        output = {
            name + "_rank_statistic": (vals[i] < param_priors[name]).sum()
            for i, name in enumerate(names)
        }
        for name, p_prior in param_priors.items():
            output[name + "_ground_truth"] = p_prior

        # Compute posterior sensitivities
        for name, val in zip(names, vals):
            output[name + "_mean"] = np.mean(val)
            output[name + "_sd"] = np.std(val)
            output[name + "_z_score"] = (
                output[name + "_mean"] - output[name + "_ground_truth"]
            ) / output[name + "_sd"]
            output[name + "_shrinkage"] = 1 - (output[name + "_sd"] / prior_sd[name]) ** 2
            output[name + "_ESS"] = (
                diagnostics["ess"]
                .loc[diagnostics["ess"]["parameter"] == name, "ESS"]
                .values[0]
            )
            output[name + "_ESS_per_iter"] = (
                diagnostics["ess"]
                .loc[diagnostics["ess"]["parameter"] == name, "ESS_per_iter"]
                .values[0]
            )
            output[name + "_tail_ESS"] = (
                diagnostics["ess"]
                .loc[diagnostics["ess"]["parameter"] == name, "tail_ESS"]
                .values[0]
            )
            output[name + "_tail_ESS_per_iter"] = (
                diagnostics["ess"]
                .loc[diagnostics["ess"]["parameter"] == name, "tail_ESS_per_iter"]
                .values[0]
            )
            output[name + "_Rhat"] = (
                diagnostics["rhat"]
                .loc[diagnostics["rhat"]["parameter"] == name, "Rhat"]
                .values[0]
            )

        output["n_bad_ebfmi"] = diagnostics["e_bfmi"]["problematic"].sum()
        output["n_divergences"] = int(diagnostics["n_divergences"])
        output["n_max_treedepth"] = int(diagnostics["n_max_treedepth"])
        output["warning_code"] = warning_code
    else:
        # Convert output to Numpy array
        names, _ = _xarray_to_ndarray(prior_sample.prior, var_names=var_names)

        # Generate output dictionary
        output = {
            name + "_rank_statistic": np.nan
            for i, name in enumerate(names)
        }
        for name, p_prior in param_priors.items():
            output[name + "_ground_truth"] = p_prior

        # Posterior sensitivities
        for name in names:
            output[name + "_mean"] = np.nan
            output[name + "_sd"] = np.nan
            output[name + "_z_score"] = np.nan
            output[name + "_shrinkage"] = np.nan
            output[name + "_ESS"] = np.nan
            output[name + "_ESS_per_iter"] = np.nan
            output[name + "_tail_ESS"] = np.nan
            output[name + "_tail_ESS_per_iter"] = np.nan
            output[name + "_Rhat"] = np.nan

        output["n_bad_ebfmi"] = np.nan
        output["n_divergences"] = np.nan
        output["n_max_treedepth"] = np.nan
        output["warning_code"] = np.nan
    output['error'] = err_str
    output['L'] = L

    # Convert output dict to tidied format (multiple rows, one per parameter)
    tidy_df = _tidy_single_sbc_result(output, trial)

    # Use lock for thread-safe file writing
    with write_lock:
        # Append tidied rows to CSV (headers already written at start)
        tidy_df.to_csv(
            output_csv,
            mode='a',
            header=False,
            index=False
        )

    return None


def _get_prior_sds(
    prior_predictive_model,
    prior_predictive_model_data,
    var_names,
    measured_data,
    n_prior_draws_for_sd,
    samples_dir,
    remove_sample_files,
):
    """Compute standard deviations of prior parameters."""
    with disable_logging():
        output_dir = _get_output_dir(samples_dir, "prior_pred")
        prior_samples_cmdstanpy = prior_predictive_model.sample(
            data=prior_predictive_model_data,
            fixed_param=True,
            iter_sampling=n_prior_draws_for_sd,
            chains=1,
            show_progress=False,
            output_dir=output_dir,
        )
    prior_samples = az.from_cmdstanpy(
        prior=prior_samples_cmdstanpy, prior_predictive=measured_data
    )

    if remove_sample_files:
        _remove_sampling_files(prior_samples_cmdstanpy.runset)

    # Compute prior sd's
    names, vals = _xarray_to_ndarray(prior_samples.prior)
    prior_sd = {
        name: np.std(val)
        for name, val in zip(names, vals)
        if name in var_names or _base_name(name) in var_names
    }

    return prior_sd


def _base_name(name):
    if "[" not in name or name[-1] != "]":
        return name

    ind = name.rfind("[")
    comma_inds = [ind + i for i, char in enumerate(name[ind:]) if char == ","]

    if len(comma_inds) == 0:
        if ind == len(name) - 2 or not name[ind + 1 : -1].isnumeric():
            return name
        else:
            return name[:ind]
    elif comma_inds[0] == ind + 1 or comma_inds[-1] == len(name) - 2:
        return name
    else:
        if name[ind + 1 : -1].replace(",", "").isnumeric():
            return name[:ind]
        else:
            return name


def _tidy_single_sbc_result(output_dict, trial_number):
    """Tidy a single SBC result dict into multiple rows (one per parameter).

    Parameters
    ----------
    output_dict : dict
        Single result dictionary from _perform_sbc
    trial_number : int
        Trial number for this result

    Returns
    -------
    df : pandas DataFrame
        Tidied output with one row per parameter
    """
    # Extract parameter names from rank_statistic keys
    rank_stat_keys = [k for k in output_dict.keys() if k.endswith("_rank_statistic")]
    params = [k[:-len("_rank_statistic")] for k in rank_stat_keys]

    stats = [
        "ground_truth",
        "rank_statistic",
        "mean",
        "sd",
        "shrinkage",
        "z_score",
        "Rhat",
        "ESS",
        "ESS_per_iter",
        "tail_ESS",
        "tail_ESS_per_iter",
    ]

    aux_keys = ["n_divergences", "n_bad_ebfmi", "n_max_treedepth", "warning_code", "L", "error"]

    rows = []
    for param in params:
        row = {"trial": trial_number, "parameter": param}

        # Add statistics for this parameter
        for stat in stats:
            key = f"{param}_{stat}"
            row[stat] = output_dict.get(key, np.nan)

        # Add auxiliary columns
        for aux_key in aux_keys:
            row[aux_key] = output_dict.get(aux_key, np.nan)

        rows.append(row)

    return pd.DataFrame(rows)


def _tidy_sbc_output(sbc_output):
    """Tidy output from sbc().

    Returns
    -------
    output : DataFrame
        Tidy data frame with SBC results.

    """
    df = sbc_output.copy()
    df["trial"] = df.index.values

    rank_stat_cols = list(df.columns[df.columns.str.contains("_rank_statistic")])
    params = [col[: col.rfind("_rank_statistic")] for col in rank_stat_cols]

    dfs = []
    stats = [
        "ground_truth",
        "rank_statistic",
        "mean",
        "sd",
        "shrinkage",
        "z_score",
        "Rhat",
        "ESS",
        "ESS_per_iter",
        "tail_ESS",
        "tail_ESS_per_iter",
    ]

    aux_cols = [
        "n_divergences",
        "n_bad_ebfmi",
        "n_max_treedepth",
        "warning_code",
        "L",
        "trial",
        "error"
    ]
    for param in params:
        cols = [param + "_" + stat for stat in stats]
        sub_df = df[cols + aux_cols].rename(
            columns={old_col: new_col for old_col, new_col in zip(cols, stats)}
        )
        sub_df["parameter"] = param
        dfs.append(sub_df)

    return pd.concat(dfs, ignore_index=True)


def _xarray_to_ndarray(ds, var_names=None, omit_dunders=True):
    """Convert xarray data set with coordinates `chain` and `draw` to a
    Numpy array and a list of row labels for the Numpy array.

    Parameters
    ----------
    ds : xarray Dataset
        Dataset stored in an ArviZ InferenceData instance. This is
        usually contained in the `prior` or `posterior` fields of the
        InferenceData instance.
    var_names : list of strings, default None
        Names of variables to include. If None, include all. Note that
        for multidimensional variables, `var_names` only has the root
        name. E.g., `var_names=['x', 'y']`, *not* something like
        `var_names=['x[0]', 'x[1]', 'y[0,0]']`.
    omit_dunders : bool, default True
        If True, omit any variable that ends in '__' unless it is
        explicitly included in `var_names`. These are not variables, but
        sampling information computed by Stan.

    Output
    ------
    names : list of strings
        List of names of headings. For a multidimensional xarray
        DataArray, each entry in the array is the variable name,
        followed indexing information. As an example, if `A` is a 2x2
        matrix, then there will be names 'A[0,0]', 'A[0,1]', 'A[1,0]',
        and 'A[1,1]'.
    vals : 2D Numpy array
        Each row has the samples for a given variable. If combined is
        True, each row is a concatenation of samples from all chains.

    """

    names, vals = arviz.sel_utils.xarray_to_ndarray(
        ds, var_names=var_names, combined=True
    )

    names = [
        name.replace("\n", "[").replace(", ", ",") + "]" if "\n" in name else name
        for name in names
    ]

    if var_names is None:
        var_names = []

    inds = [
        i
        for i, name in enumerate(names)
        if name in var_names or len(name) < 2 or name[-2:] != "__"
    ]

    return [name for i, name in enumerate(names) if i in inds], vals[inds, :]


def _ebfmi(energy):
    """Compute energy-Bayes fraction of missing information"""
    return np.sum(np.diff(energy) ** 2) / (len(energy) - 1) / np.var(energy)


def _get_var_name(parameter):
    """Convert a parameter name to a var_name. Example: 'alpha[0,1]'
    returns 'alpha'."""
    if parameter[-1] != "]":
        return parameter

    ind = parameter.rfind("[")
    if ind == 0 or ind == len(parameter) - 1:
        return parameter

    substr = parameter[ind + 1 : -1]
    if len(substr) == 0:
        return parameter

    if not substr[0].isdigit():
        return parameter

    if not substr[-1].isdigit():
        return parameter

    for char in substr:
        if not (char.isdigit() or char == ","):
            return parameter

    if ",," in substr:
        return parameter

    return parameter[:ind]


def _parameters_to_arviz_var_names(samples, parameters):
    """Convert individual parameter names to ArviZ var_names.
    E.g., 'alpha[0]' becomes 'alpha'.
    """
    if parameters is not None:
        if type(parameters) not in (list, tuple):
            raise RuntimeError("`parameters` must be a list or tuple.")

        sample_vars = list(samples.posterior.data_vars)

        var_names = [
            param if param in sample_vars else _get_var_name(param)
            for param in parameters
        ]

        # Drop any duplicates
        var_names_set = set()
        var_names = [
            var_name
            for var_name in var_names
            if not (var_name in var_names_set or var_names_set.add(var_name))
        ]

        for var_name in var_names:
            if var_name not in sample_vars:
                raise RuntimeError(f"variable {var_name} not in the input.")
    else:
        var_names = None

    return var_names


def _samples_parameters_to_df(samples, parameters, omit=[]):
    """Convert ArviZ InferenceData posterior results to a data frame"""
    if parameters is None:
        params = None
    else:
        param_dict = {}
        for param in parameters:
            if type(param) in [tuple, list]:
                param_dict[param[0]] = param[1]
            else:
                param_dict[param] = param

        params = list(param_dict.keys())
        parameters = [(k, v) for k, v in param_dict.items()]

    var_names_arviz = _parameters_to_arviz_var_names(samples, params)

    diagnostics = ('diverging',) if hasattr(samples.sample_stats, 'diverging') else tuple()

    df = arviz_to_dataframe(samples, var_names=var_names_arviz, diagnostics=diagnostics, df_package='pandas')

    if parameters is None:
        parameters = [col for col in df.columns if _screen_param(col, omit)]
        params = copy.copy(parameters)

    if hasattr(samples.sample_stats, 'diverging'):
        cols = list(params) + ["chain__", "draw__", "diverging__"]
    else:
        cols = list(params) + ["chain__", "draw__"]

    return parameters, df[cols].copy()


def _screen_param(param, omit):
    if param in ["chain__", "draw__", "diverging__"]:
        return False

    for pattern in omit:
        if type(pattern) == re.Pattern:
            if bool(pattern.match(param)):
                return False
        elif pattern == param:
            return False

    return True


def _get_output_dir(samples_dir, model_name):
    """Hand-specify directory for sampling results. This is necessary
    so that CmdStanPy doesn't try to access files with the same name
    while doing multiprocessing."""
    # Set up time stamp + random nine-digit number
    now_ns = time.time_ns()
    dt = datetime.datetime.fromtimestamp(now_ns // 1000000000)
    s = dt.strftime("%Y%m%d%H%M%S") + "." + str(int(now_ns % 1000000000)).zfill(9)
    s += "-" + str(random.randint(0, 1000000000)).zfill(9)

    subdir = model_name + "-" + s
    return os.path.join(samples_dir, subdir)


def _remove_sampling_files(runset):
    """Remove files laying around from sampling run."""

    #### Alternatively, we could delete files one-by-one, in case other
    #### stuff is in the directory we're deleting from. Here, we
    #### obliterate everything in the directory.

    # files = []
    # try:
    #     files.append(runset._csv_files)
    # except:
    #     pass

    # try:
    #     files.append(runset._stdout_files)
    # except:
    #     pass

    # try:
    #     files.append(runset._stderr_files)
    # except:
    #     pass

    # try:
    #     files.append(runset._diagnostic_files)
    # except:
    #     pass

    # try:
    #     files.append(runset._profile_files)
    # except:
    #     pass

    # for flist in files:
    #     for fname in flist:
    #         try:
    #             os.remove(fname)
    #         except:
    #             pass
    #
    # try:
    #     os.rmdir(runset._output_dir)
    # except:
    #     pass

    try:
        shutil.rmtree(runset._output_dir)
    except:
        pass


@contextlib.contextmanager
def disable_logging(level=logging.CRITICAL):
    """Context manager for disabling logging when doing MCMC sampling.

    Parameters
    ----------
    level : int, default logging.CRITICAL

    """
    previous_level = logging.root.manager.disable

    logging.disable(level)

    try:
        yield
    finally:
        logging.disable(previous_level)

